import logging
import builtins

from picle.models import Outputters
from pydantic import (
    BaseModel,
    StrictBool,
    StrictInt,
    StrictStr,
    Field,
)
from typing import Union, Optional, List
from ..common import log_error_or_result, listen_events
from ..nornir.nornir_picle_shell_common import NornirCommonArgs, NorniHostsFilters
from .netbox_picle_shell_common import NetboxClientRunJobArgs
from norfab.workers.netbox_worker.netbox_models import NetboxCommonArgs

log = logging.getLogger(__name__)


class SyncDeviceFactsDatasourcesNornir(NornirCommonArgs, NorniHostsFilters):
    @staticmethod
    def run(*args, **kwargs):
        kwargs["datasource"] = "nornir"
        return SyncDeviceFactsCommand.run(*args, **kwargs)

    class PicleConfig:
        outputter = Outputters.outputter_nested


class UpdateDeviceFactsDatasources(BaseModel):
    nornir: SyncDeviceFactsDatasourcesNornir = Field(
        None,
        description="Use Nornir service to retrieve data from devices",
    )


class SyncDeviceFactsCommand(NetboxCommonArgs, NetboxClientRunJobArgs):
    dry_run: Optional[StrictBool] = Field(
        None,
        description="Return information that would be pushed to Netbox but do not push it",
        json_schema_extra={"presence": True},
        alias="dry-run",
    )
    devices: Union[List[StrictStr], StrictStr] = Field(
        None,
        description="List of Netbox devices to sync",
    )
    batch_size: StrictInt = Field(
        10, description="Number of devices to process at a time", alias="batch-size"
    )
    datasource: UpdateDeviceFactsDatasources = Field(
        "nornir",
        description="Service to use to retrieve device data",
    )
    branch: StrictStr = Field(None, description="Branching plugin branch name to use")

    @staticmethod
    @listen_events
    def run(uuid, **kwargs):
        NFCLIENT = builtins.NFCLIENT
        workers = kwargs.pop("workers", "any")
        timeout = kwargs.pop("timeout", 600)
        kwargs["timeout"] = timeout * 0.9
        verbose_result = kwargs.pop("verbose_result", False)
        nowait = kwargs.pop("nowait", False)

        if isinstance(kwargs.get("devices"), str):
            kwargs["devices"] = [kwargs["devices"]]

        result = NFCLIENT.run_job(
            "netbox",
            "sync_device_facts",
            workers=workers,
            kwargs=kwargs,
            timeout=timeout,
            uuid=uuid,
            nowait=nowait,
        )

        if nowait:
            return result, Outputters.outputter_nested

        result = log_error_or_result(result, verbose_result=verbose_result)

        return result

    class PicleConfig:
        outputter = Outputters.outputter_nested


class SyncDeviceInterfacesDatasourcesNornir(NornirCommonArgs, NorniHostsFilters):
    @staticmethod
    def run(*args, **kwargs):
        kwargs["datasource"] = "nornir"
        return SyncDeviceInterfacesCommand.run(*args, **kwargs)

    class PicleConfig:
        outputter = Outputters.outputter_nested


class SyncDeviceInterfacesDatasources(BaseModel):
    nornir: SyncDeviceInterfacesDatasourcesNornir = Field(
        None,
        description="Use Nornir service to retrieve data from devices",
    )


class SyncDeviceInterfacesCommand(NetboxCommonArgs, NetboxClientRunJobArgs):
    dry_run: Optional[StrictBool] = Field(
        None,
        description="Return information that would be pushed to Netbox but do not push it",
        json_schema_extra={"presence": True},
        alias="dry-run",
    )
    devices: Union[List[StrictStr], StrictStr] = Field(
        None,
        description="List of Netbox devices to sync",
    )
    datasource: SyncDeviceInterfacesDatasources = Field(
        "nornir",
        description="Service to use to retrieve device data",
    )
    batch_size: StrictInt = Field(
        10, description="Number of devices to process at a time", alias="batch-size"
    )
    branch: StrictStr = Field(None, description="Branching plugin branch name to use")

    @staticmethod
    @listen_events
    def run(uuid, **kwargs):
        NFCLIENT = builtins.NFCLIENT
        workers = kwargs.pop("workers", "any")
        timeout = kwargs.pop("timeout", 600)
        kwargs["timeout"] = timeout * 0.9
        verbose_result = kwargs.pop("verbose_result", False)
        nowait = kwargs.pop("nowait", False)

        if isinstance(kwargs.get("devices"), str):
            kwargs["devices"] = [kwargs["devices"]]

        result = NFCLIENT.run_job(
            "netbox",
            "sync_device_interfaces",
            workers=workers,
            kwargs=kwargs,
            timeout=timeout,
            uuid=uuid,
            nowait=nowait,
        )

        if nowait:
            return result, Outputters.outputter_nested

        result = log_error_or_result(result, verbose_result=verbose_result)

        return result

    class PicleConfig:
        outputter = Outputters.outputter_nested


class SyncDeviceIPAddressesDatasourcesNornir(NornirCommonArgs, NorniHostsFilters):
    @staticmethod
    def run(*args, **kwargs):
        kwargs["datasource"] = "nornir"
        return SyncDeviceIPAddressesCommand.run(*args, **kwargs)

    class PicleConfig:
        outputter = Outputters.outputter_nested


class SyncDeviceIPAddressesDatasources(BaseModel):
    nornir: SyncDeviceIPAddressesDatasourcesNornir = Field(
        None,
        description="Use Nornir service to retrieve data from devices",
    )


class SyncDeviceIPAddressesCommand(NetboxCommonArgs, NetboxClientRunJobArgs):
    dry_run: Optional[StrictBool] = Field(
        None,
        description="Return information that would be pushed to Netbox but do not push it",
        json_schema_extra={"presence": True},
        alias="dry-run",
    )
    devices: Union[List[StrictStr], StrictStr] = Field(
        None,
        description="List of Netbox devices to sync",
    )
    datasource: SyncDeviceIPAddressesDatasources = Field(
        "nornir",
        description="Service to use to retrieve device data",
    )
    batch_size: StrictInt = Field(
        10, description="Number of devices to process at a time", alias="batch-size"
    )
    branch: StrictStr = Field(None, description="Branching plugin branch name to use")

    @staticmethod
    @listen_events
    def run(uuid, **kwargs):
        NFCLIENT = builtins.NFCLIENT
        workers = kwargs.pop("workers", "any")
        timeout = kwargs.pop("timeout", 600)
        kwargs["timeout"] = timeout * 0.9
        verbose_result = kwargs.pop("verbose_result", False)
        nowait = kwargs.pop("nowait", False)

        if isinstance(kwargs.get("devices"), str):
            kwargs["devices"] = [kwargs["devices"]]

        result = NFCLIENT.run_job(
            "netbox",
            "sync_device_ip",
            workers=workers,
            kwargs=kwargs,
            timeout=timeout,
            uuid=uuid,
            nowait=nowait,
        )

        if nowait:
            return result, Outputters.outputter_nested

        result = log_error_or_result(result, verbose_result=verbose_result)

        return result

    class PicleConfig:
        outputter = Outputters.outputter_nested


class SyncDeviceCommands(BaseModel):
    facts: SyncDeviceFactsCommand = Field(
        None,
        description="Sync device facts e.g. serial number",
    )
    interfaces: SyncDeviceInterfacesCommand = Field(
        None,
        description="Sync device interfaces",
    )
    ip_addresses: SyncDeviceIPAddressesCommand = Field(
        None, description="Sync device interface IP addresses", alias="ip-addresses"
    )



import logging
import builtins

from rich.console import Console
from picle.models import PipeFunctionsModel, Outputters
from ..netbox.netbox_picle_shell_get_containerlab_inventory import (
    NetboxDeviceFilters,
    GetContainerlabInventoryCommand,
)
from pydantic import (
    StrictBool,
    Field,
)
from typing import Optional
from ..common import log_error_or_result, listen_events

RICHCONSOLE = Console()
SERVICE = "containerlab"
log = logging.getLogger(__name__)


# ---------------------------------------------------------------------------------------------
# CONTAINERLAB DEPLOY NETBOX COMMANDS MODELS
# ---------------------------------------------------------------------------------------------


class DeployNetboxDeviceFilters(NetboxDeviceFilters):
    @staticmethod
    def run(*args, **kwargs):
        filters = {
            k: kwargs.pop(k)
            for k in [
                "tenant",
                "q",
                "model",
                "platform",
                "region",
                "role",
                "site",
                "status",
                "tag",
            ]
            if k in kwargs
        }
        # need to be a list
        kwargs["filters"] = [filters]
        return DeployNetboxCommand.run(*args, **kwargs)

    class PicleConfig:
        outputter = Outputters.outputter_nested
        pipe = PipeFunctionsModel


class DeployNetboxCommand(GetContainerlabInventoryCommand):
    reconfigure: StrictBool = Field(
        False,
        description="Destroy the lab and then re-deploy it.",
        json_schema_extra={"presence": True},
    )
    filters: DeployNetboxDeviceFilters = Field(
        None, description="Netbox device filters to generate lab inventory for"
    )
    dry_run: Optional[StrictBool] = Field(
        None,
        description="Do not deploy, only fetch inventory from Netbox",
        json_schema_extra={"presence": True},
        alias="dry-run",
    )

    @staticmethod
    def source_lab_name():
        NFCLIENT = builtins.NFCLIENT
        ret = []
        result = NFCLIENT.run_job("containerlab", "get_running_labs")
        for wname, wres in result.items():
            ret.extend(wres["result"])
        return ret

    @staticmethod
    @listen_events
    def run(uuid, *args, **kwargs):
        NFCLIENT = builtins.NFCLIENT
        verbose_result = kwargs.pop("verbose_result")
        workers = kwargs.pop("workers", "any")

        if not any(k in kwargs for k in ["devices", "filters", "tenant"]):
            raise ValueError(
                "Devices list or Netbox filters or Tenant name must be provided."
            )

        if not any(k in kwargs for k in ["lab_name", "tenant"]):
            raise ValueError("Lab name or Tenant name must be provided.")

        if kwargs.get("devices"):
            if not isinstance(kwargs.get("devices"), list):
                kwargs["devices"] = [kwargs["devices"]]

        result = NFCLIENT.run_job(
            "containerlab",
            "deploy_netbox",
            workers=workers,
            kwargs=kwargs,
            args=args,
            uuid=uuid,
        )

        return log_error_or_result(
            result, verbose_result=verbose_result, verbose_on_fail=True
        )

    class PicleConfig:
        outputter = Outputters.outputter_nested
        pipe = PipeFunctionsModel



import logging
import pprint
import yaml

from norfab.core.nfapi import NorFab
from robot.api.deco import keyword
from robot.api import ContinuableFailure
from robot.api import logger
from nornir_salt.plugins.functions import TabulateFormatter

log = logging.getLogger(__name__)

# Global vars
DATA = {}


def clean_global_data():
    global DATA
    DATA.clear()


class NorFabRobot:
    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_AUTO_KEYWORDS = False
    ROBOT_LIBRARY_DOC_FORMAT = "reST"  # reStructuredText
    ROBOT_LISTENER_API_VERSION = 3

    def __init__(
        self,
        inventory="./inventory.yaml",
        log_level="WARNING",
    ):
        self.ROBOT_LIBRARY_LISTENER = self

        # initiate NorFab
        self.nf = NorFab(inventory=inventory, log_level=log_level)
        self.nf.start()
        self.client = self.nf.make_client()

    def start_suite(self, data, result):
        pass

    def end_suite(self, data, result):
        print("NorFab ROBOT - Exiting")
        self.nf.destroy()

    @keyword("Workers")
    def workers(self, *args, **kwargs):
        """Collect workers to target"""
        if args:
            DATA["workers"] = args
        else:
            DATA["workers"] = kwargs.pop("workers", "all")

    @keyword("Hosts")
    def hosts(self, *args, **kwargs):
        """Collect hosts to target"""
        if args:
            DATA["hosts"] = {"FB": ", ".join(args), **kwargs}
        else:
            DATA["hosts"] = kwargs

    @keyword("nr.test")
    def nr_test(self, *args, **kwargs):
        """Run nr.test  task"""
        tests_pass = 0
        tests_fail = 0
        tests_results = []
        commands_output = {}
        if args:
            kwargs["suite"] = args[0]
        kwargs = {
            **DATA.pop("hosts", {"FB": "*"}),
            **kwargs,
            "remove_tasks": False,
            "add_details": True,
            "return_tests_suite": True,
            "to_dict": False,
        }
        logger.info(f"Running nr.test with kwargs '{kwargs}', global DATA '{DATA}'")
        has_errors = False
        # run this function
        ret = self.client.run_job(
            service="nornir",
            task="test",
            workers=DATA.get("workers", "all"),
            kwargs=kwargs,
        )
        # iterate over results and log tests and task statuses
        for worker, worker_results in ret.items():
            for result in worker_results["result"]["test_results"]:
                host = result["host"]
                # evaluate and log test result
                if "success" in result:
                    if (
                        result["failed"]
                        or result["exception"]
                        or not result["success"]
                        or "traceback" in str(result["result"]).lower()
                    ):
                        tests_fail += 1
                        has_errors = True
                        logger.error(
                            (
                                f'{worker} worker, {host} test "{result["name"]}" - '
                                f'<span style="background-color: #CE3E01">"{result["exception"]}"</span>'
                            ),
                            html=True,
                        )
                    else:
                        tests_pass += 1
                        logger.info(
                            (
                                f'{worker} worker, {host} test "{result["name"]}" - '
                                f'<span style="background-color: #97BD61">success</span>'
                            ),
                            html=True,
                        )
                    # save test results to log them later
                    tests_results.append({"worker": worker, **result})
                # evaluate and log task result
                else:
                    # log exception for task
                    if result["failed"] or result["exception"]:
                        has_errors = True
                        logger.error(
                            (
                                f'{worker} worker, {host} task "{result["name"]}" - '
                                f'<span style="background-color: #CE3E01">"{result["exception"].strip()}"</span>'
                            ),
                            html=True,
                        )
                    # save device commands output to log it later
                    commands_output.setdefault(host, {})
                    commands_output[host][result["name"]] = result["result"]
        # clear global state to prep for next test
        clean_global_data()

        tests_results_html_table = TabulateFormatter(
            tests_results,
            tabulate={"tablefmt": "html"},
            headers=[
                "worker",
                "host",
                "name",
                "result",
                "failed",
                "task",
                "test",
                "criteria",
                "exception",
            ],
        )

        tests_results_csv_table = [
            f'''"{i['worker']}","{i['host']}","{i['name']}","{i['result']}","{i['failed']}","{i['task']}","{i['test']}","{i['criteria']}","{i['exception']}"'''
            for i in tests_results
        ]
        tests_results_csv_table.insert(
            0,
            '"worker","host","name","result","failed","task","test","criteria","exception"',
        )
        tests_results_csv_table = "\n".join(tests_results_csv_table)

        # form nested HTML of commands output
        devices_output_html = []
        for host in sorted(commands_output.keys()):
            commands = commands_output[host]
            commands_output_html = []
            for command, result in commands.items():
                commands_output_html.append(
                    f'<p><details style="margin-left:20px;"><summary>{command}</summary><p style="margin-left:20px;"><font face="courier new">{result}</font></p></details></p>'
                )
            devices_output_html.append(
                f'<p><details><summary>{host} ({len(commands_output_html)} commands)</summary><p>{"".join(commands_output_html)}</p></details></p>'
            )

        # form nested HTML for devices tes suite
        devices_test_suite = []
        for worker, worker_results in ret.items():
            for host in sorted(worker_results["result"]["suite"].keys()):
                suite_content = worker_results["result"]["suite"][host]
                devices_test_suite.append(
                    f'<p><details><summary>{host} ({len(suite_content)} tests)</summary><p style="margin-left:20px;">{yaml.dump(suite_content, default_flow_style=False)}</p></details></p>'
                )

        logger.info(
            f"<details><summary>Workers results</summary>{pprint.pformat(ret)}</details>",
            html=True,
        )
        logger.info(
            f"<details><summary>Test suite results details</summary><p>{tests_results_html_table}</p></details>",
            html=True,
        )
        logger.info(
            f"<details><summary>Test suite results CSV table</summary><p>{tests_results_csv_table}</p></details>",
            html=True,
        )
        logger.info(
            f"<details><summary>Devices tests suites content</summary>{''.join(devices_test_suite)}</details>",
            html=True,
        )
        logger.info(
            f"<details><summary>Collected devices output</summary>{''.join(devices_output_html)}</details>",
            html=True,
        )
        logger.info(
            (
                f"Tests completed - {tests_pass + tests_fail}, "
                f'<span style="background-color: #97BD61">success - {tests_pass}</span>, '
                f'<span style="background-color: #CE3E01">failed - {tests_fail}</span>'
            ),
            html=True,
        )

        # raise if has errors
        if has_errors:
            raise ContinuableFailure("Tests failed")
        # return test results with no errors in structured format
        return ret

    @keyword("nr.cli")
    def nr_cli(self, *args, **kwargs):
        """Run Nornir service cli task"""
        log.info(
            f"Running nr.cli with args '{args}', kwargs '{kwargs}', global DATA '{DATA}'"
        )
        has_errors = False
        if args:
            kwargs["commands"] = args
        kwargs = {
            **DATA.pop("hosts", {"FB": "*"}),
            **kwargs,
            "add_details": True,
            "to_dict": False,
        }
        # run this function
        ret = self.client.run_job(
            service="nornir",
            task="cli",
            workers=DATA.get("workers", "all"),
            kwargs=kwargs,
        )
        # extract results for the host
        for worker, worker_results in ret.items():
            for result in worker_results["result"]:
                host = result["host"]
                # evaluate and log results
                if (
                    result["failed"]
                    or result["exception"]
                    or "traceback" in str(result["result"]).lower()
                ):
                    has_errors = True
                    logger.error(
                        (
                            f'<details><summary>{worker} worker, {host} device, command "{result["name"]}" failed - '
                            f'<span style="background-color: #CE3E01">"{result["exception"]}"</span></summary>'
                            f'<p style="margin-left:20px;"><font face="courier new">{result["result"]}'
                            f"</font></p></details>"
                        ),
                        html=True,
                    )
                else:
                    logger.info(
                        (
                            f'<details><summary>{worker} worker, {host} device, command "{result["name"]}" - '
                            f'<span style="background-color: #97BD61">success</span></summary>'
                            f'<p style="margin-left:20px;"><font face="courier new">{result["result"]}'
                            f"</font></p></details>"
                        ),
                        html=True,
                    )
        logger.info(
            f"<details><summary>Workers results</summary>{pprint.pformat(ret)}</details>",
            html=True,
        )
        # clean global state to prep for next test
        clean_global_data()
        # raise exception if cli command failed
        if has_errors:
            raise ContinuableFailure(ret)
        # return ret with no errors in structured format
        return ret

    @keyword("nr.cfg")
    def nr_cfg(self, *args, **kwargs):
        """Run Nornir service cfg task"""
        log.info(
            f"Running nr.cfg with args '{args}', kwargs '{kwargs}', global DATA '{DATA}'"
        )
        if args:
            kwargs["config"] = args
        kwargs = {
            **DATA.pop("hosts", {"FB": "*"}),
            **kwargs,
            "add_details": True,
            "to_dict": False,
        }
        has_errors = False
        # run this function
        ret = self.client.run_job(
            service="nornir",
            task="cfg",
            workers=DATA.get("workers", "all"),
            kwargs=kwargs,
        )
        # extract results for the host
        for worker, worker_results in ret.items():
            for result in worker_results["result"]:
                host = result["host"]
                # evaluate and log results
                if (
                    result["failed"]
                    or result["exception"]
                    or "traceback" in str(result["result"]).lower()
                ):
                    has_errors = True
                    logger.error(
                        (
                            f'<details><summary>{worker} worker, {host} device, "{result["name"]}" failed - '
                            f'<span style="background-color: #CE3E01">"{result["exception"]}"</span></summary>'
                            f'<p style="margin-left:20px;"><font face="courier new">{result["result"]}'
                            f"</font></p></details>"
                        ),
                        html=True,
                    )
                else:
                    logger.info(
                        (
                            f'<details><summary>{worker} worker, {host} device, "{result["name"]}" - '
                            f'<span style="background-color: #97BD61">success</span></summary>'
                            f'<p style="margin-left:20px;"><font face="courier new">{result["result"]}'
                            f"</font></p></details>"
                        ),
                        html=True,
                    )
        logger.info(
            f"<details><summary>Workers results</summary>{pprint.pformat(ret)}</details>",
            html=True,
        )
        # clean global state to prep for next test
        clean_global_data()
        # raise exception if cli command failed
        if has_errors:
            raise ContinuableFailure(ret)
        # return ret with no errors in structured format
        return ret



import logging
import sqlite3
import zlib
import zmq
import time
import orjson
import os
import threading
import queue
import hashlib
import glob
import shutil
from contextlib import contextmanager
from uuid import uuid4  # random uuid

from .security import generate_certificates
from . import NFP
from norfab.core.inventory import NorFabInventory
from norfab.utils.markdown_results import markdown_results
from typing import Any, Optional, Tuple, Dict, List, Set, Union

log = logging.getLogger(__name__)

# --------------------------------------------------------------------------------------------
# NIRFAB client, credits to https://rfc.zeromq.org/spec/9/
# --------------------------------------------------------------------------------------------


# Job status constants
class JobStatus:
    NEW = "NEW"
    SUBMITTING = "SUBMITTING"  # POST sent, waiting for broker ACK
    DISPATCHED = "DISPATCHED"  # Broker dispatched to workers
    STARTED = "STARTED"  # At least one worker started processing
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    STALE = "STALE"  # Job exceeded deadline without completion


class ClientJobDatabase:
    """Lightweight client-side job and events store."""

    def __init__(self, db_path: str, jobs_compress: bool = True):
        self.db_path = db_path
        self.jobs_compress = jobs_compress
        self._local = threading.local()
        self._lock = threading.Lock()
        self._initialize_database()

    def _get_connection(self) -> sqlite3.Connection:
        if not hasattr(self._local, "conn"):
            self._local.conn = sqlite3.connect(
                self.db_path, check_same_thread=False, timeout=30.0
            )
            self._local.conn.row_factory = sqlite3.Row
            # Enable WAL mode for better concurrent access
            self._local.conn.execute("PRAGMA journal_mode=WAL")
            self._local.conn.execute("PRAGMA synchronous=NORMAL")
            # performance PRagmas
            self._local.conn.execute("PRAGMA cache_size=5000")  # 5MB cache
            self._local.conn.execute("PRAGMA temp_store=MEMORY")  # Use RAM for temp
            self._local.conn.execute("PRAGMA mmap_size=30000000")  # Memory-mapped I/O
            self._local.conn.execute("PRAGMA query_only=False")  # Default
            self._local.conn.execute("PRAGMA busy_timeout=5000")  # 5s timeout

        return self._local.conn

    @contextmanager
    def _transaction(self, write: bool = False):
        conn = self._get_connection()
        if write:
            with self._lock:
                try:
                    yield conn
                    conn.commit()
                except Exception:
                    conn.rollback()
                    raise
        else:
            yield conn

    def _compress(self, data: Dict | List | Any) -> bytes:
        raw = orjson.dumps(data)
        return zlib.compress(raw, level=6) if self.jobs_compress else raw

    def _decompress(self, payload: bytes | None) -> Any:
        if payload is None:
            return None
        return orjson.loads(zlib.decompress(payload) if self.jobs_compress else payload)

    def _initialize_database(self) -> None:
        with self._transaction(write=True) as conn:
            conn.execute("""
                CREATE TABLE IF NOT EXISTS jobs (
                    uuid TEXT PRIMARY KEY,
                    service TEXT NOT NULL,
                    task TEXT NOT NULL,
                    args BLOB,
                    kwargs BLOB,
                    timeout INTEGER,
                    deadline REAL,
                    status TEXT DEFAULT 'NEW',
                    workers_requested TEXT,
                    workers_dispatched TEXT,
                    workers_started TEXT,
                    workers_completed TEXT,
                    result_data BLOB,
                    errors TEXT,
                    received_timestamp TEXT NOT NULL,
                    started_timestamp TEXT,
                    completed_timestamp TEXT,
                    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                    last_poll_timestamp REAL DEFAULT 0
                )
                """)
            conn.execute("""
                CREATE TABLE IF NOT EXISTS events (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    job_uuid TEXT NOT NULL,
                    message TEXT NOT NULL,
                    severity TEXT DEFAULT 'INFO',
                    task TEXT,
                    event_data BLOB,
                    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                    FOREIGN KEY (job_uuid) REFERENCES jobs(uuid) ON DELETE CASCADE
                )
                """)
            conn.execute("CREATE INDEX IF NOT EXISTS idx_jobs_status ON jobs(status)")
            conn.execute(
                "CREATE INDEX IF NOT EXISTS idx_jobs_created ON jobs(created_at)"
            )
            conn.execute("CREATE INDEX IF NOT EXISTS idx_jobs_service ON jobs(service)")
            conn.execute(
                "CREATE INDEX IF NOT EXISTS idx_jobs_deadline ON jobs(deadline)"
            )
            conn.execute(
                "CREATE INDEX IF NOT EXISTS idx_events_job_uuid ON events(job_uuid)"
            )
            conn.execute(
                "CREATE INDEX IF NOT EXISTS idx_events_severity ON events(severity)"
            )
            conn.execute(
                "CREATE INDEX IF NOT EXISTS idx_jobs_last_poll ON jobs(last_poll_timestamp)"
            )

    def add_job(
        self,
        uuid: str,
        service: str,
        task: str,
        workers: Any,
        args: list,
        kwargs: dict,
        timeout: int,
        deadline: float,
    ) -> None:

        with self._transaction(write=True) as conn:
            conn.execute(
                """
                INSERT INTO jobs (uuid, service, task, args, kwargs, timeout, deadline,
                                  status, workers_requested, received_timestamp)
                VALUES (?, ?, ?, ?, ?, ?, ?, 'NEW', ?, ?)
                """,
                (
                    uuid,
                    service,
                    task,
                    self._compress({"args": args or []}),
                    self._compress({"kwargs": kwargs or {}}),
                    timeout,
                    deadline,
                    orjson.dumps(workers).decode("utf-8"),
                    time.ctime(),
                ),
            )

    def update_job(
        self,
        uuid: str,
        *,
        status: str | None = None,
        workers_dispatched: Set[str] | List[str] | None = None,
        workers_started: Set[str] | List[str] | None = None,
        workers_completed: Set[str] | List[str] | None = None,
        result_data: dict | None = None,
        errors: List[str] | None = None,
        append_errors: List[str] | None = None,
        started_ts: str | None = None,
        completed_ts: str | None = None,
        last_poll_ts: float | None = None,
    ) -> None:
        fields = []
        values: List[Any] = []

        def _store_set(label: str, value: Set[str] | List[str] | None):
            if value is None:
                return
            if isinstance(value, set):
                value = sorted(value)
            fields.append(f"{label} = ?")
            values.append(orjson.dumps(value).decode("utf-8"))

        if status:
            fields.append("status = ?")
            values.append(status)
        _store_set("workers_dispatched", workers_dispatched)
        _store_set("workers_started", workers_started)
        _store_set("workers_completed", workers_completed)
        if result_data is not None:
            fields.append("result_data = ?")
            values.append(self._compress(result_data))
        if errors is not None:
            fields.append("errors = ?")
            values.append(orjson.dumps(errors).decode("utf-8"))
        if started_ts:
            fields.append("started_timestamp = ?")
            values.append(started_ts)
        if completed_ts:
            fields.append("completed_timestamp = ?")
            values.append(completed_ts)
        if last_poll_ts is not None:
            fields.append("last_poll_timestamp = ?")
            values.append(last_poll_ts)

        if not fields and not append_errors:
            return

        # Handle appending errors to existing errors
        if append_errors:
            with self._transaction(write=False) as conn:
                cur = conn.execute("SELECT errors FROM jobs WHERE uuid = ?", (uuid,))
                row = cur.fetchone()
                existing_errors = (
                    orjson.loads(row["errors"]) if row and row["errors"] else []
                )
            existing_errors.extend(append_errors)
            fields.append("errors = ?")
            values.append(orjson.dumps(existing_errors).decode("utf-8"))

        if not fields:
            return

        with self._transaction(write=True) as conn:
            conn.execute(
                f"UPDATE jobs SET {', '.join(fields)} WHERE uuid = ?",
                (*values, uuid),
            )

    def fetch_jobs(
        self,
        statuses: List[str] = None,
        limit: int = 10,
        min_poll_age: float = 0,
        service: str = None,
        task: str = None,
        workers_completed: List[str] = None,
        last: int = None,
    ) -> List[dict]:
        """Fetch jobs with flexible filtering and complete job attributes.

        Args:
            statuses: List of job statuses to filter by (default: all statuses)
            limit: Maximum number of jobs to return (used when last is not specified)
            min_poll_age: Minimum seconds since last poll (for rate limiting GET requests)
            service: Service name to filter by (optional)
            task: Task name to filter by (optional)
            workers_completed: List of worker names that completed the job (optional)
            last: Return only the last x number of jobs (newest first), overrides limit (optional)

        Returns:
            List of job dictionaries with complete attributes including:
            uuid, service, task, args, kwargs, workers_*, status, result_data,
            errors, timestamps, etc.
        """
        conditions = []
        params = []

        # Filter by statuses
        if statuses:
            placeholders = ",".join(["?"] * len(statuses))
            conditions.append(f"status IN ({placeholders})")
            params.extend(statuses)

        # Filter by poll age (for dispatcher throttling)
        if min_poll_age > 0:
            poll_threshold = time.time() - min_poll_age
            conditions.append(
                "(last_poll_timestamp IS NULL OR last_poll_timestamp <= ?)"
            )
            params.append(poll_threshold)

        # Filter by service
        if service:
            conditions.append("service = ?")
            params.append(service)

        # Filter by task
        if task:
            conditions.append("task = ?")
            params.append(task)

        # Filter by workers_completed (JSON contains check)
        if workers_completed:
            # For each worker, check if it's in the JSON array
            worker_conditions = []
            for worker in workers_completed:
                # SQLite JSON string contains check
                worker_conditions.append("workers_completed LIKE ?")
                params.append(f'%"{worker}"%')
            conditions.append(f"({' OR '.join(worker_conditions)})")

        # Build WHERE clause
        where_clause = " AND ".join(conditions) if conditions else "1=1"

        # Determine order and limit based on 'last' parameter
        order_direction = "DESC" if last is not None else "ASC"
        result_limit = last if last is not None else limit

        with self._transaction(write=False) as conn:
            cur = conn.execute(
                f"""
                SELECT uuid, service, task, args, kwargs, workers_requested, timeout, deadline,
                       workers_dispatched, workers_started, workers_completed, status,
                       result_data, errors, received_timestamp, started_timestamp,
                       completed_timestamp, created_at, last_poll_timestamp
                FROM jobs
                WHERE {where_clause}
                ORDER BY created_at {order_direction}
                LIMIT ?
                """,
                (*params, result_limit),
            )
            rows = cur.fetchall()
        return [self._hydrate(row) for row in rows]

    def get_job(self, uuid: str) -> dict | None:
        with self._transaction(write=False) as conn:
            cur = conn.execute(
                """
                  SELECT uuid, service, task, args, kwargs, timeout, deadline, status,
                       workers_requested, workers_dispatched, workers_started,
                       workers_completed, result_data, errors, created_at, completed_timestamp,
                       last_poll_timestamp
                FROM jobs WHERE uuid = ?
                """,
                (uuid,),
            )
            row = cur.fetchone()
        return self._hydrate(row) if row else None

    def _hydrate(self, row: sqlite3.Row) -> dict:
        if row is None:
            return None
        data = dict(row)
        data["args"] = self._decompress(data.get("args")) or {"args": []}
        data["kwargs"] = self._decompress(data.get("kwargs")) or {"kwargs": {}}
        data["args"] = data["args"].get("args", [])
        data["kwargs"] = data["kwargs"].get("kwargs", {})
        for field in [
            "workers_requested",
            "workers_dispatched",
            "workers_started",
            "workers_completed",
        ]:
            if data.get(field):
                data[field] = orjson.loads(data[field])
            else:
                data[field] = []
        if data.get("result_data"):
            data["result_data"] = self._decompress(data["result_data"])
        if data.get("errors"):
            data["errors"] = orjson.loads(data["errors"])
        else:
            data["errors"] = []
        data["timeout"] = data.get("timeout", 600) or 600
        data["deadline"] = data.get("deadline", 0) or 0
        data["last_poll_timestamp"] = data.get("last_poll_timestamp", 0) or 0
        return data

    def add_event(
        self,
        job_uuid: str,
        message: str,
        severity: str = "INFO",
        task: str | None = None,
        event_data: dict | None = None,
    ) -> None:
        with self._transaction(write=True) as conn:
            conn.execute(
                """
                INSERT INTO events (job_uuid, message, severity, task, event_data)
                VALUES (?, ?, ?, ?, ?)
                """,
                (
                    job_uuid,
                    message,
                    severity,
                    task,
                    self._compress(event_data or {}),
                ),
            )

    def close(self) -> None:
        if hasattr(self._local, "conn"):
            self._local.conn.close()
            delattr(self._local, "conn")

    def jobs_stats(self) -> dict:
        """Return job-level statistics from the jobs database."""
        stats = {
            # job counts
            "total_jobs": 0,  # Total number of jobs in the database
            "total_events": 0,  # Total number of events in the database
            "jobs_by_status": {},  # Job counts keyed by status string
            "jobs_by_service": {},  # Job counts keyed by service name
            "events_by_severity": {},  # Event counts keyed by severity level
            # job timing
            "jobs_last_24h": 0,  # Jobs created in the last 24 hours
            "oldest_job_ts": None,  # received_timestamp of the oldest job
            "newest_job_ts": None,  # received_timestamp of the newest job
            "avg_completion_seconds": None,  # Average completion time in seconds for COMPLETED jobs
        }

        with self._transaction(write=False) as conn:
            # ── Job counts ────────────────────────────────────────────────
            stats["total_jobs"] = conn.execute("SELECT COUNT(*) FROM jobs").fetchone()[
                0
            ]

            cur = conn.execute("SELECT status, COUNT(*) FROM jobs GROUP BY status")
            stats["jobs_by_status"] = {row[0]: row[1] for row in cur.fetchall()}

            cur = conn.execute("SELECT service, COUNT(*) FROM jobs GROUP BY service")
            stats["jobs_by_service"] = {row[0]: row[1] for row in cur.fetchall()}

            stats["total_events"] = conn.execute(
                "SELECT COUNT(*) FROM events"
            ).fetchone()[0]

            cur = conn.execute(
                "SELECT severity, COUNT(*) FROM events GROUP BY severity"
            )
            stats["events_by_severity"] = {row[0]: row[1] for row in cur.fetchall()}

            # ── Job timing ────────────────────────────────────────────────
            cutoff = time.strftime(
                "%Y-%m-%d %H:%M:%S", time.localtime(time.time() - 86400)
            )
            stats["jobs_last_24h"] = conn.execute(
                "SELECT COUNT(*) FROM jobs WHERE created_at >= ?", (cutoff,)
            ).fetchone()[0]

            row = conn.execute(
                "SELECT MIN(received_timestamp), MAX(received_timestamp) FROM jobs"
            ).fetchone()
            if row:
                stats["oldest_job_ts"] = row[0]
                stats["newest_job_ts"] = row[1]

            # Average completion time for jobs that have both timestamps recorded.
            # Timestamps are stored as ctime() strings, so we parse them in Python.
            cur = conn.execute("""
                SELECT received_timestamp, completed_timestamp
                FROM jobs
                WHERE status = 'COMPLETED'
                  AND received_timestamp IS NOT NULL
                  AND completed_timestamp IS NOT NULL
                """)
            completed_rows = cur.fetchall()
            if completed_rows:
                import calendar

                durations = []
                for r in completed_rows:
                    try:
                        start = calendar.timegm(time.strptime(r[0]))
                        end = calendar.timegm(time.strptime(r[1]))
                        durations.append(end - start)
                    except (ValueError, OverflowError):
                        pass
                if durations:
                    stats["avg_completion_seconds"] = round(
                        sum(durations) / len(durations), 3
                    )

        return stats

    def jobs_db_stats(self) -> dict:
        """Return SQLite database-level statistics."""
        stats = {
            # file sizes
            "db_size_bytes": 0,  # Database file size on disk in bytes
            "db_wal_size_bytes": 0,  # WAL file size on disk in bytes (0 if absent)
            "db_shm_size_bytes": 0,  # SHM file size on disk in bytes (0 if absent)
            "db_total_disk_bytes": 0,  # Sum of db + wal + shm file sizes
            # page / space
            "db_page_size": 0,  # SQLite page size in bytes
            "db_page_count": 0,  # Total number of pages allocated in the database
            "db_max_page_count": 0,  # Maximum allowed number of pages
            "db_freelist_count": 0,  # Number of unused (free) pages
            "db_used_size_bytes": 0,  # Approximate used space = (page_count - freelist_count) * page_size
            # configuration / runtime
            "db_cache_size": 0,  # Current cache size setting (number of pages)
            "db_journal_mode": "",  # Journal mode (e.g. 'wal', 'delete')
            "db_wal_autocheckpoint": 0,  # WAL auto-checkpoint page threshold (0 = disabled)
            "db_mmap_size": 0,  # Memory-mapped I/O size in bytes
            "db_soft_heap_limit": 0,  # SQLite soft heap memory limit in bytes (0 = none)
            "db_hard_heap_limit": 0,  # SQLite hard heap memory limit in bytes (0 = none)
            "db_data_version": 0,  # Monotonic counter incremented on each external write
            "db_sqlite_version": sqlite3.sqlite_version,  # SQLite library version string
        }

        # File sizes on disk (main db, WAL, SHM)
        for attr, suffix in [
            ("db_size_bytes", ""),
            ("db_wal_size_bytes", "-wal"),
            ("db_shm_size_bytes", "-shm"),
        ]:
            path = self.db_path + suffix
            try:
                stats[attr] = os.path.getsize(path) if os.path.exists(path) else 0
            except OSError:
                stats[attr] = 0
        stats["db_total_disk_bytes"] = (
            stats["db_size_bytes"]
            + stats["db_wal_size_bytes"]
            + stats["db_shm_size_bytes"]
        )

        with self._transaction(write=False) as conn:
            stats["db_page_size"] = conn.execute("PRAGMA page_size").fetchone()[0]
            stats["db_page_count"] = conn.execute("PRAGMA page_count").fetchone()[0]
            stats["db_max_page_count"] = conn.execute(
                "PRAGMA max_page_count"
            ).fetchone()[0]
            stats["db_freelist_count"] = conn.execute(
                "PRAGMA freelist_count"
            ).fetchone()[0]
            stats["db_cache_size"] = conn.execute("PRAGMA cache_size").fetchone()[0]
            stats["db_journal_mode"] = conn.execute("PRAGMA journal_mode").fetchone()[0]
            stats["db_wal_autocheckpoint"] = conn.execute(
                "PRAGMA wal_autocheckpoint"
            ).fetchone()[0]
            stats["db_mmap_size"] = conn.execute("PRAGMA mmap_size").fetchone()[0]
            stats["db_data_version"] = conn.execute("PRAGMA data_version").fetchone()[0]
            try:
                stats["db_soft_heap_limit"] = conn.execute(
                    "PRAGMA soft_heap_limit"
                ).fetchone()[0]
            except Exception:
                pass
            try:
                stats["db_hard_heap_limit"] = conn.execute(
                    "PRAGMA hard_heap_limit"
                ).fetchone()[0]
            except Exception:
                pass

        page_size = stats["db_page_size"]
        used_pages = stats["db_page_count"] - stats["db_freelist_count"]
        stats["db_used_size_bytes"] = max(0, used_pages) * page_size

        return stats


def recv(client):
    """
    Receiver thread: processes all incoming messages from the broker and updates the database.

    This function continuously polls the client's broker socket for messages
    until the client's exit event is set. It handles:
    - EVENT messages: stored in the events table
    - RESPONSE messages: updates job status in the database based on response type

    The receiver thread is the ONLY thread that reads from the socket, eliminating
    contention issues. All job state changes are persisted to the database.

    Args:
        client (object): The client instance containing the broker socket,
                         poller, job_db, and configuration.
    """
    while not client.exit_event.is_set() and not client.destroy_event.is_set():
        # Poll socket for messages every 500ms interval
        try:
            items = client.poller.poll(500)
        except KeyboardInterrupt:
            break
        except Exception:
            continue

        if not items:
            continue

        with client.socket_lock:
            try:
                msg = client.broker_socket.recv_multipart(zmq.NOBLOCK)
            except zmq.Again:
                continue

        client.stats_recv_from_broker += 1

        # Message format: [empty, header, command, service, uuid, status, payload]
        if len(msg) < 7:
            log.error(f"{client.name} - received malformed message: {msg}")
            continue

        command = msg[2]
        juuid = msg[4].decode("utf-8")
        status = msg[5].decode("utf-8")

        log.debug(
            f"{client.name} - received '{command}' message from broker, juuid {juuid}, status {status}"
        )

        if command == NFP.STREAM:
            payload = msg[6]  # payload is a chunk of bytes
            handle_stream(client, juuid, status, payload)
            continue

        try:
            payload = orjson.loads(msg[6])
        except Exception as e:
            log.error(
                f"{client.name} - failed to parse message, error '{e}'", exc_info=True
            )
            continue

        # Handle EVENT messages
        if command == NFP.EVENT:
            handle_event(client, juuid, payload, msg)

        # Handle RESPONSE messages
        if command == NFP.RESPONSE:
            handle_response(client, juuid, status, payload)

        # handle MMI messages
        if command == NFP.MMI:
            client.mmi_queue.put(msg)


def handle_event(client: object, juuid: str, payload: dict, msg: list):
    """
    Handle EVENT messages and update job database accordingly.

    Args:
        client: The client instance
        juuid: Job UUID
        payload: Event payload dictionary
        msg: Original message multipart for queue
    """
    client.event_queue.put(msg)
    client.stats_recv_event_from_broker += 1
    client.job_db.add_event(
        job_uuid=juuid,
        message=payload.get("message", ""),
        severity=payload.get("severity", "INFO"),
        task=payload.get("task"),
        event_data=payload,
    )


def handle_response(client, juuid: str, status: str, payload: dict):
    """
    Handle RESPONSE messages and update job database accordingly.

    Uses job status to determine context:
    - SUBMITTING: expecting broker 202 with workers list
    - DISPATCHED/STARTED: expecting worker ACKs (202), results (200), or pending (300)

    Status codes:
    - 202: Accepted (POST acknowledged by broker or worker)
    - 200: OK (GET completed with results)
    - 300: Pending (job still in progress)
    - 4xx: Client errors
    - 5xx: Server errors
    """
    job = client.job_db.get_job(juuid)
    if not job:
        log.debug(f"{client.name} - received response for unknown job {juuid}")
        return

    # Broker accepted POST - contains dispatched workers list
    if status == "202":  # ACCEPTED
        workers_list = payload["workers"]
        client.job_db.update_job(
            juuid,
            status=JobStatus.DISPATCHED,
            workers_dispatched=workers_list,
            started_ts=time.ctime(),
        )
        log.debug(f"{client.name} - job {juuid} dispatched to workers: {workers_list}")
        return

    # Worker created the job
    if status == "201":  # JOB CREATED
        worker_single = payload["worker"]
        started = set(job.get("workers_started", []))
        started.add(worker_single)
        client.job_db.update_job(
            juuid,
            status=JobStatus.STARTED,
            workers_started=list(started),
        )
        log.debug(
            f"{client.name} - job {juuid} acknowledged by worker: {worker_single}"
        )
        return

        # GET dispatched to workers (broker 202 response to GET)
        if workers_list:
            log.debug(
                f"{client.name} - job {juuid} GET dispatched to workers: {workers_list}"
            )
        return

    # Handle 200 OK - GET completed with results
    if status == "200":
        dispatched = set(job.get("workers_dispatched", []))
        completed = set(job.get("workers_completed", []))
        existing_results = job.get("result_data") or {}

        # Merge new results with existing (results keyed by worker name)
        if isinstance(payload, dict):
            for worker_name in payload.keys():
                completed.add(worker_name)
            existing_results.update(payload)

        is_complete = completed == dispatched and len(dispatched) > 0

        client.job_db.update_job(
            juuid,
            status=JobStatus.COMPLETED if is_complete else JobStatus.STARTED,
            workers_completed=list(completed),
            result_data=existing_results,
            completed_ts=time.ctime() if is_complete else None,
        )

        if is_complete:
            log.debug(f"{client.name} - job {juuid} completed")
        return

    # Handle 300 Pending - job still in progress
    if status == "300":
        worker = payload.get("worker")
        if worker and worker not in job["workers_started"]:
            job["workers_started"].append(worker)
            client.job_db.update_job(
                juuid,
                status=JobStatus.STARTED,
                workers_started=job["workers_started"],
            )
        return

    # Handle error statuses (4xx, 5xx)
    if status.startswith("4") or status.startswith("5"):
        error_msg = payload.get("error", payload.get("status", f"Error {status}"))
        client.job_db.update_job(
            juuid,
            status=JobStatus.FAILED,
            append_errors=[error_msg],
            completed_ts=time.ctime(),
        )
        log.error(f"{client.name} - job {juuid} failed: {error_msg}")
        return


def handle_stream(client, juuid: str, status: str, payload: bytes):
    job = client.job_db.get_job(juuid)
    file_transfer = client.file_transfers.get(juuid)

    if not job:
        log.error(f"{client.name} - received stream for unknown job {juuid}")
        return

    if not file_transfer:
        log.error(f"{client.name} - received stream for unknown file transfer {juuid}")
        return

    destination = file_transfer["destination"]  # file object
    size = len(payload)
    file_transfer["credit"] += 1  # Up to PIPELINE chunks in transit
    file_transfer["total_bytes_received"] += size

    # save received chunk
    destination.write(payload)
    file_transfer["file_hash"].update(payload)

    # check if done
    if file_transfer["total_bytes_received"] >= file_transfer["size_bytes"]:
        destination.close()

        # check md5hash mismatch
        file_hash = file_transfer["file_hash"].hexdigest()
        if file_hash != file_transfer["md5hash"]:
            client.job_db.update_job(
                juuid,
                status=JobStatus.FAILED,
                errors=["Download failed, MD5 hash mismatch"],
            )
            log.error(
                f"{client.name} - file download failed, MD5 hash mismatch, job '{juuid}', filename '{destination.name}'"
            )

        log.debug(
            f"{client.name} - finished file download, job '{juuid}', filename '{destination.name}'"
        )
        return

    # request next set of chunks up to credit
    while file_transfer["credit"] > 0 and file_transfer["chunk_requests_remaining"] > 0:
        file_transfer["offset"] += file_transfer[
            "chunk_size"
        ]  # Offset of next chunk request
        service = client.ensure_bytes(job["service"])
        uuid_bytes = client.ensure_bytes(juuid)
        workers = client.ensure_bytes(job["workers_requested"])
        request = client.ensure_bytes(
            {
                "offset": file_transfer["offset"],
            }
        )
        client.send_to_broker(NFP.PUT, service, workers, uuid_bytes, request)
        file_transfer["credit"] -= 1
        file_transfer["chunk_requests_remaining"] -= 1


def dispatch_new_jobs(client):
    """
    Find NEW jobs and send POST requests to broker.
    Non-blocking: sends request and updates status to SUBMITTING.
    """
    for job in client.job_db.fetch_jobs(
        [JobStatus.NEW], limit=client.dispatch_batch_size
    ):
        juuid = job["uuid"]

        try:
            # Send POST request (non-blocking)
            service = client.ensure_bytes(job["service"])
            uuid_bytes = client.ensure_bytes(juuid)
            workers = client.ensure_bytes(job["workers_requested"])
            request = client.ensure_bytes(
                {
                    "task": job["task"],
                    "kwargs": job["kwargs"] or {},
                    "args": job["args"] or [],
                }
            )

            client.send_to_broker(NFP.POST, service, workers, uuid_bytes, request)

            # Update status - receiver will handle the response
            client.job_db.update_job(
                juuid,
                status=JobStatus.SUBMITTING,
                last_poll_ts=time.time(),
            )
            log.debug(f"{client.name} - dispatched POST for job {juuid}")

        except Exception as e:
            msg = f"{client.name} - failed to dispatch job {juuid}: {e}"
            log.error(msg, exc_info=True)
            client.job_db.update_job(
                juuid,
                status=JobStatus.FAILED,
                errors=[msg],
                completed_ts=time.ctime(),
            )


def poll_active_jobs(client):
    """
    Find active jobs and send GET requests to poll for results.
    Non-blocking: sends request with 5-second throttling via last_poll_timestamp.
    """
    # Jobs that are ready for GET polling (dispatched or started)
    active_statuses = [JobStatus.DISPATCHED, JobStatus.STARTED]

    # fetch_jobs filters by min_poll_age to enforce polling throttle
    for job in client.job_db.fetch_jobs(
        active_statuses,
        limit=client.dispatch_batch_size,
        min_poll_age=client.poll_interval,
    ):
        juuid = job["uuid"]
        deadline = job["deadline"]
        now = time.time()

        # Check if job has exceeded deadline
        if now >= deadline:
            client.job_db.update_job(
                juuid,
                status=JobStatus.STALE,
                errors=["Job deadline reached without completion"],
                completed_ts=time.ctime(),
            )
            continue

        try:
            # Send GET request (non-blocking)
            service = client.ensure_bytes(job["service"])
            uuid_bytes = client.ensure_bytes(juuid)
            workers = client.ensure_bytes(job["workers_dispatched"])
            request = client.ensure_bytes(
                {
                    "task": job["task"],
                    "kwargs": job["kwargs"] or {},
                    "args": job["args"] or [],
                }
            )

            client.send_to_broker(NFP.GET, service, workers, uuid_bytes, request)

            # Update last_poll_ts to enforce 5-second throttle
            client.job_db.update_job(
                juuid,
                last_poll_ts=time.time(),
            )
            log.debug(f"{client.name} - sent GET poll for job {juuid}")

        except Exception as e:
            log.error(f"{client.name} - failed to poll job {juuid}: {e}", exc_info=True)
            # Don't fail the job on poll error, just log and retry next cycle


def dispatcher(client):
    """
    Dispatcher thread: sends POST and GET requests asynchronously.

    This thread:
    1. Finds NEW jobs and sends POST requests to broker
    2. Finds DISPATCHED/STARTED jobs and sends GET requests to poll for results

    It does NOT wait for responses - the receiver thread handles all incoming
    messages and updates the database.

    Args:
        client (object): The client instance containing job_db, exit_event, and configuration.
    """
    while not client.exit_event.is_set() and not client.destroy_event.is_set():
        try:
            dispatch_new_jobs(client)
            poll_active_jobs(client)
        except Exception as e:
            log.error(f"{client.name} - dispatcher error: {e}", exc_info=True)
        time.sleep(0.1)


class NFPClient(object):
    """
    NFPClient is a client class for interacting with a broker using ZeroMQ for messaging.
    It handles sending and receiving messages, managing connections, and performing tasks.

    Attributes:
        broker (str): The broker address.
        ctx (zmq.Context): The ZeroMQ context.
        broker_socket (zmq.Socket): The ZeroMQ socket for communication with the broker.
        poller (zmq.Poller): The ZeroMQ poller for managing socket events.
        name (str): The name of the client.
        stats_send_to_broker (int): Counter for messages sent to the broker.
        stats_recv_from_broker (int): Counter for messages received from the broker.
        stats_reconnect_to_broker (int): Counter for reconnections to the broker.
        stats_recv_event_from_broker (int): Counter for events received from the broker.
        client_private_key_file (str): Path to the client's private key file.
        broker_public_key_file (str): Path to the broker's public key file.

    Methods:
        __init__(inventory, broker, name, exit_event=None, event_queue=None):
            Initializes the NFPClient instance with the given parameters.
        ensure_bytes(workers) -> bytes:
            Helper function to convert workers target to bytes.
        reconnect_to_broker():
            Connects or reconnects to the broker.
        send_to_broker(command, service, workers, uuid, request):
            Sends a message to the broker.
        rcv_from_broker(command, service, uuid):
            Waits for a response from the broker.
        post(service, task, args=None, kwargs=None, workers="all", uuid=None, timeout=600):
            Sends a job request to the broker and returns the result.
        get(service, task=None, args=None, kwargs=None, workers="all", uuid=None, timeout=600):
            Sends a job reply message to the broker requesting job results.
        get_iter(service, task, args=None, kwargs=None, workers="all", uuid=None, timeout=600):
            Sends a job reply message to the broker requesting job results and yields results iteratively.
        fetch_file(url, destination=None, chunk_size=250000, pipeline=10, timeout=600, read=False):
            Downloads a file from the Broker File Sharing Service.
        run_job(service, task, uuid=None, args=None, kwargs=None, workers="all", timeout=600):
            Runs a job and returns results produced by workers.
        run_job_iter(service, task, uuid=None, args=None, kwargs=None, workers="all", timeout=600):
            Runs a job and yields results produced by workers iteratively.
        destroy():
            Cleans up and destroys the client instance.

    Args:
        inventory (NorFabInventory): The inventory object containing base directory information.
        broker: The broker object for communication.
        name (str): The name of the client.
        exit_event (threading.Event, optional): An event to signal client exit. Defaults to None.
        event_queue (queue.Queue, optional): A queue for handling events. Defaults to None.
    """

    broker = None
    ctx = None
    broker_socket = None
    poller = None
    name = None
    stats_send_to_broker = 0
    stats_recv_from_broker = 0
    stats_reconnect_to_broker = 0
    stats_recv_event_from_broker = 0
    client_private_key_file = None
    broker_public_key_file = None
    public_keys_dir = None
    private_keys_dir = None

    def __init__(
        self,
        inventory: NorFabInventory,
        broker: str,
        name: str,
        exit_event: Optional[threading.Event] = None,
        event_queue: Optional[queue.Queue] = None,
    ):
        self.inventory = inventory
        self.name = name
        self.zmq_name = f"{self.name}-{uuid4().hex}"
        self.broker = broker
        self.base_dir = os.path.join(
            self.inventory.base_dir, "__norfab__", "files", "client", self.name
        )
        self.file_transfers = {}  # file transfers tracker
        self.zmq_auth = self.inventory.broker.get("zmq_auth", True)
        self.socket_lock = threading.Lock()  # used to protect socket object
        self.build_message = NFP.MessageBuilder()

        # create base directories
        os.makedirs(self.base_dir, exist_ok=True)

        self.job_db = ClientJobDatabase(
            os.path.join(self.base_dir, f"{self.name}.db"),
            jobs_compress=True,
        )

        # generate certificates and create directories
        if self.zmq_auth is not False:
            generate_certificates(
                self.base_dir,
                cert_name=self.name,
                broker_keys_dir=os.path.join(
                    self.inventory.base_dir,
                    "__norfab__",
                    "files",
                    "broker",
                    "public_keys",
                ),
                inventory=self.inventory,
            )
            self.public_keys_dir = os.path.join(self.base_dir, "public_keys")
            self.private_keys_dir = os.path.join(self.base_dir, "private_keys")

        self.ctx = zmq.Context()
        self.poller = zmq.Poller()
        self.reconnect_to_broker()

        self.exit_event = threading.Event() if exit_event is None else exit_event
        self.destroy_event = (
            threading.Event()
        )  # destroy event, used by worker to stop its client
        self.mmi_queue = queue.Queue(maxsize=0)
        self.event_queue = event_queue or queue.Queue(maxsize=1000)

        # Configuration for dispatcher
        self.poll_interval = 0.5  # Seconds between GET polls for same job (throttling)
        self.dispatch_batch_size = 10  # Max jobs to process per dispatch cycle

        # start receiver thread - handles all incoming messages
        self.recv_thread = threading.Thread(
            target=recv, daemon=True, name=f"{self.name}_recv_thread", args=(self,)
        )
        self.recv_thread.start()

        # start dispatcher thread - sends POST/GET requests asynchronously
        self.dispatcher_thread = threading.Thread(
            target=dispatcher,
            daemon=True,
            name=f"{self.name}_dispatcher",
            args=(self,),
        )
        self.dispatcher_thread.start()

    def ensure_bytes(self, value: Any) -> bytes:
        """
        Helper function to convert value to bytes.
        """
        if isinstance(value, bytes):
            return value
        # transform string to bytes
        if isinstance(value, str):
            return value.encode("utf-8")
        # convert value to json string
        else:
            return orjson.dumps(value)

    def reconnect_to_broker(self):
        """
        Connect or reconnect to the broker.

        This method handles the connection or reconnection to the broker by:

        - Closing the existing broker socket if it exists.
        - Creating a new DEALER socket.
        - Setting the socket options including the identity and linger.
        - Loading the client's private and public keys for CURVE encryption.
        - Loading the broker's public key for CURVE encryption.
        - Connecting the socket to the broker.
        - Registering the socket with the poller for incoming messages.
        - Logging the connection status.
        - Incrementing the reconnect statistics counter.
        """
        if self.broker_socket:
            self.poller.unregister(self.broker_socket)
            self.broker_socket.close()

        self.broker_socket = self.ctx.socket(zmq.DEALER)
        self.broker_socket.setsockopt_unicode(zmq.IDENTITY, self.zmq_name, "utf8")
        self.broker_socket.linger = 0

        if self.zmq_auth is not False:
            # We need two certificates, one for the client and one for
            # the server. The client must know the server's public key
            # to make a CURVE connection.
            self.client_private_key_file = os.path.join(
                self.private_keys_dir, f"{self.name}.key_secret"
            )
            client_public, client_secret = zmq.auth.load_certificate(
                self.client_private_key_file
            )
            self.broker_socket.curve_secretkey = client_secret
            self.broker_socket.curve_publickey = client_public

            # The client must know the server's public key to make a CURVE connection.
            self.broker_public_key_file = os.path.join(
                self.public_keys_dir, "broker.key"
            )
            server_public, _ = zmq.auth.load_certificate(self.broker_public_key_file)
            self.broker_socket.curve_serverkey = server_public

        self.broker_socket.connect(self.broker)
        self.poller.register(self.broker_socket, zmq.POLLIN)
        log.debug(f"{self.name} - client connected to broker at '{self.broker}'")
        self.stats_reconnect_to_broker += 1

    def send_to_broker(self, command, service, workers, uuid, request):
        """
        Sends a command to the broker.

        Args:
            command (str): The command to send (e.g., NFP.POST, NFP.GET).
            service (str): The service to which the command is related.
            workers (str): The workers involved in the command.
            uuid (str): The unique identifier for the request.
            request (str): The request payload to be sent.
        """
        if command == NFP.POST:
            msg = self.build_message.client_to_broker_post(
                command=command,
                service=service,
                workers=workers,
                uuid=uuid,
                request=request,
            )
        elif command == NFP.GET:
            msg = self.build_message.client_to_broker_get(
                command=command,
                service=service,
                workers=workers,
                uuid=uuid,
                request=request,
            )
        elif command == NFP.PUT:
            msg = self.build_message.client_to_broker_put(
                command=command,
                service=service,
                workers=workers,
                uuid=uuid,
                request=request,
            )
        elif command == NFP.MMI:
            msg = self.build_message.client_to_broker_mmi(
                command=command,
                service=service,
                workers=workers,
                uuid=uuid,
                request=request,
            )
        else:
            log.error(
                f"{self.name} - cannot send '{command}' to broker, command unsupported"
            )
            return

        log.debug(f"{self.name} - sending '{msg}'")

        with self.socket_lock:
            if self.destroy_event.is_set():
                return
            self.broker_socket.send_multipart(msg)
            self.stats_send_to_broker += 1

    def mmi(
        self,
        service: str,
        task: str = None,
        args: list = None,
        kwargs: dict = None,
        workers: Union[str, list] = "all",
        uuid: hex = None,
        timeout: int = 30,
    ) -> dict:
        """
        Send an MMI (management interface) request to a service via the broker.

        MMI requests are intended for lightweight, broker-routed management or
        introspection operations (e.g., service metadata, health, inventory-like
        queries) that return a single aggregated response payload.

        Args:
            service: Target service name.
            task: Service task name to execute.
            args: Positional arguments for the task.
            kwargs: Keyword arguments for the task.
            workers: Workers selector. Can be ``"all"``, ``"any"``, or a list of names.
            uuid: Optional request UUID. If not provided, a random UUID is generated.
            timeout: Maximum time (seconds) to wait for the MMI reply.

        Returns:
            Dictionary containing ``status``, ``results``, and ``errors`` keys:

                - ``status``: HTTP-like status code as a string (e.g., ``"200"``, ``"408"``).
                - ``results``: Decoded JSON payload from the broker/service.
                - ``errors``: List of error strings.
        """
        service_str = service
        uuid_str = uuid or uuid4().hex
        args = args or []
        kwargs = kwargs or {}
        ret = {"status": "200", "results": {}, "errors": []}

        service = self.ensure_bytes(service_str)
        uuid = self.ensure_bytes(uuid_str)
        workers = self.ensure_bytes(workers)

        request = self.ensure_bytes(
            {"task": task, "kwargs": kwargs or {}, "args": args or []}
        )

        self.send_to_broker(NFP.MMI, service, workers, uuid, request)

        deadline = time.time() + timeout
        while time.time() < deadline:
            # check if need to stop
            if self.exit_event.is_set() or self.destroy_event.is_set():
                ret["errors"].append(
                    f"{self.name} - '{uuid_str}:{service_str}' MMI interrupted (client stopping)"
                )
                ret["status"] = "499"
                break

            try:
                msg = self.mmi_queue.get(block=True, timeout=0.5)
                self.mmi_queue.task_done()
            except queue.Empty:
                continue

            (
                empty,
                reply_header,
                reply_command,
                reply_service,
                reply_uuid,
                reply_status,
                reply_task_result,
            ) = msg

            # Defer unrelated replies and continue scanning.
            if reply_uuid != uuid:
                self.mmi_queue.put(msg)
                continue

            if reply_header != NFP.BROKER and reply_command != NFP.MMI:
                ret["errors"].append(
                    f"{self.name} - '{uuid_str}:{service_str}' MMI unexpected reply header/command"
                )
                ret["status"] = reply_status.decode("utf-8")
                break

            try:
                ret["results"] = orjson.loads(reply_task_result)
                ret["status"] = reply_status.decode("utf-8")
            except Exception as e:
                ret["errors"].append(
                    f"{self.name} - '{uuid_str}:{service_str}' MMI failed to decode reply payload: {e}"
                )
                ret["results"] = {"status": "Invalid MMI response payload"}
                ret["status"] = "500"
            break
        else:
            msg = f"{self.name} - '{uuid_str}:{service_str}' MMI request {timeout}s timeout exceeded."
            log.error(msg)
            ret["errors"].append(msg)
            ret["results"] = {"status": "MMI Request Timeout"}
            ret["status"] = "408"

        return ret

    def delete_fetched_files(self, filepath: str = "*") -> dict:
        """
        Delete files and folders matching the filepath glob pattern.

        Args:
            filepath (str): Glob pattern to match files/folders. Default is "*" (all files).

        Returns:
            dict: Dictionary with 'deleted' list of deleted paths and 'errors' list of error messages.
        """
        files_folder = os.path.join(self.base_dir, "fetchedfiles")

        result = {"deleted": [], "errors": []}

        # Build full pattern path
        pattern = os.path.join(files_folder, filepath)

        # Find all matching files and folders
        matches = glob.glob(pattern, recursive=True)

        # Sort by depth (deepest first) to avoid deleting parent before children
        matches.sort(key=lambda x: x.count(os.sep), reverse=True)

        for match in matches:
            try:
                if os.path.isfile(match):
                    os.remove(match)
                    result["deleted"].append(match)
                    log.debug(f"{self.name} - deleted file: {match}")
                elif os.path.isdir(match):
                    shutil.rmtree(match)
                    result["deleted"].append(match)
                    log.debug(f"{self.name} - deleted folder: {match}")
            except Exception as e:
                error_msg = f"Failed to delete {match}: {str(e)}"
                result["errors"].append(error_msg)
                log.error(f"{self.name} - {error_msg}")

        return result

    def fetch_file(
        self,
        url: str,
        chunk_size: int = 256000,
        pipeline: int = 10,
        timeout: int = 600,
        read: bool = False,
    ) -> Tuple[str, Any]:
        """
        Fetches a file from a given URL and saves it to a specified destination.

        Parameters:
            url (str): The URL of the file to be fetched.
            chunk_size (int, optional): The size of each chunk to be fetched. Default is 250000 bytes.
            pipeline (int, optional): The number of chunks to be fetched in transit. Default is 10.
            timeout (int, optional): The maximum time (in seconds) to wait for the file to be fetched. Default is 600 seconds.
            read (bool, optional): If True, the file content is read and returned. If False, the file path is returned. Default is False.

        Returns:
            tuple: A tuple containing the status code (str) and the reply (str). The reply can be the file content, file path, or an error message.

        Raises:
            Exception: If there is an error in fetching the file or if the file's MD5 hash does not match the expected hash.
        """

        # round up digit e.g. if 2.0 -> 2 if 2.1 -> 3 if 0.01 -> 1
        def round_up(num):
            return max(1, (int(num) + (not num.is_integer())))

        uuid = uuid4().hex
        result = {"status": "200", "content": None, "error": None}
        downloaded = False

        # run sanity checks
        if not url.startswith("nf://"):
            result["status"] = "500"
            result["error"] = "Invalid url format"
            return result

        # prevent path traversal / absolute paths
        url_path = url.replace("nf://", "")
        url_path = url_path.lstrip("/\\")
        destination = os.path.abspath(
            os.path.join(self.base_dir, "fetchedfiles", *url_path.split("/"))
        )
        fetched_root = os.path.abspath(os.path.join(self.base_dir, "fetchedfiles"))
        if os.path.commonpath([fetched_root, destination]) != fetched_root:
            result["status"] = "500"
            result["error"] = "Invalid url path"
            return result

        os.makedirs(os.path.split(destination)[0], exist_ok=True)

        self.file_transfers[uuid] = {
            "total_bytes_received": 0,  # Total bytes received
            "offset": 0,  # Offset of next chunk request
            "credit": pipeline,  # Up to PIPELINE chunks in transit
            "chunk_size": chunk_size,
            "file_hash": hashlib.md5(),
        }

        # get file details
        file_details = self.run_job(
            service="filesharing",
            workers="all",
            task="file_details",
            kwargs={"url": url},
            timeout=timeout,
        )
        for w_name, w_res in file_details.items():
            if not w_res["failed"]:
                file_details = w_res["result"]
                self.file_transfers[uuid].update(file_details)
                self.file_transfers[uuid]["w_name"] = w_name
                self.file_transfers[uuid]["chunk_requests_remaining"] = round_up(
                    file_details["size_bytes"] / chunk_size
                )
                break
        else:
            result["status"] = "404"
            result["error"] = "File download failed - file not found"
            _ = self.file_transfers.pop(uuid)
            return result

        log.debug(f"{self.name}:fetch_file - retrieved file details - {file_details}")

        # check if file already downloaded
        if os.path.isfile(destination):
            file_hash = hashlib.md5()
            with open(destination, "rb") as f:
                chunk = f.read(8192)
                while chunk:
                    file_hash.update(chunk)
                    chunk = f.read(8192)
            md5hash = file_hash.hexdigest()
            downloaded = md5hash == file_details["md5hash"]

        if file_details["exists"] and not downloaded:
            self.file_transfers[uuid]["destination"] = open(destination, "wb")
            # decrement by 1 because calling fetch_job sends first chunk
            self.file_transfers[uuid]["chunk_requests_remaining"] -= 1
            # run fetch file job
            file_fetch_job = self.run_job(
                uuid=uuid,
                service="filesharing",
                workers=[w_name],
                task="fetch_file",
                kwargs={"url": url, "offset": 0, "chunk_size": chunk_size},
                timeout=timeout,
            )
            file_fetch_job = file_fetch_job[w_name]

            if file_fetch_job["failed"]:
                result["status"] = "404"
                result["error"] = file_fetch_job["errors"]
                downloaded = False
            else:
                downloaded = True

        if downloaded:
            # Verify streaming did not mark job failed (e.g., MD5 mismatch)
            download_job = self.job_db.get_job(uuid)
            if download_job and download_job.get("status") == JobStatus.FAILED:
                result["error"] = (
                    f"File download job {uuid} failed: {download_job.get('errors', [])}"
                )
                result["status"] = "400"
            elif read:
                with open(destination, "r", encoding="utf-8") as f:
                    result["content"] = f.read()
            else:
                result["content"] = destination

        _ = self.file_transfers.pop(uuid)

        return result

    def run_job(
        self,
        service: str,
        task: str,
        uuid: str = None,
        args: list = None,
        kwargs: dict = None,
        workers: Union[str, list] = "all",
        timeout: int = 600,
        markdown: bool = False,
        nowait: bool = False,
    ) -> Any:
        """
        Run a job on the specified service and task, with optional arguments and timeout settings.

        This method submits a job to the database and waits for the dispatcher and receiver
        threads to process it asynchronously. The job progresses through states:
        NEW -> SUBMITTING -> DISPATCHED -> STARTED -> COMPLETED (or FAILED/STALE)

        Args:
            service (str): The name of the service to run the job on.
            task (str): The task to be executed.
            uuid (str, optional): A unique identifier for the job. If not provided, a new UUID will be generated. Defaults to None.
            args (list, optional): A list of positional arguments to pass to the task. Defaults to None.
            kwargs (dict, optional): A dictionary of keyword arguments to pass to the task. Defaults to None.
            workers (str, optional): The workers to run the job on. Defaults to "all".
            timeout (int, optional): The maximum time in seconds to wait for the job to complete. Defaults to 600.
            markdown (bool, optional): Convert results to markdown representation
            nowait (bool, optional): If false, wait for job to complete for timeout, return job details otherwise

        Returns:
            Any: The result of the job if successful, or None if the job failed, timed out, or became stale.
        """
        uuid = uuid or uuid4().hex
        args = args or []
        kwargs = kwargs or {}
        result = None
        job = None
        deadline = time.time() + timeout

        self.job_db.add_job(
            uuid, service, task, workers, args, kwargs, timeout, deadline
        )

        if nowait:
            return {
                "uuid": uuid,
                "service": service,
            }
        else:
            while time.time() < deadline:
                if self.exit_event.is_set() or self.destroy_event.is_set():
                    break
                job = self.job_db.get_job(uuid)
                if not job:
                    break
                if job["status"] == JobStatus.COMPLETED:
                    result = job.get("result_data")
                    break
                if job["status"] == JobStatus.FAILED:
                    log.warning(
                        f"{self.name} - job {uuid} failed: {job.get('errors', [])}"
                    )
                    break
                if job["status"] == JobStatus.STALE:
                    log.warning(
                        f"{self.name} - job {uuid} became stale: {job.get('errors', [])}"
                    )
                    break
                time.sleep(0.2)

            return markdown_results(job, service, task, kwargs) if markdown else result

    def destroy(self):
        """
        Gracefully shuts down the client.

        This method logs an interrupt message, sets the destroy event, waits
        for background threads to stop, then destroys the client context to
        ensure a clean shutdown.
        """
        log.info(f"{self.name} - client interrupt received, killing client")
        self.destroy_event.set()
        # Wait for background threads to exit before destroying the ZMQ context.
        # Without this, the dispatcher/receiver threads may still be executing
        # between their destroy_event check and a socket call, causing
        # "not a socket" ZMQError when ctx.destroy() closes sockets under them.
        if self.recv_thread.is_alive():
            self.recv_thread.join(timeout=2)
        if self.dispatcher_thread.is_alive():
            self.dispatcher_thread.join(timeout=2)
        self.job_db.close()
        self.ctx.destroy()
        # close all file transfer files
        for file_transfer in self.file_transfers.values():
            file_transfer["destination"].close()



import json
import logging
import sys
import importlib.metadata
import requests
import os
import pynetbox
import re
import time

from fnmatch import fnmatchcase
from datetime import datetime, timedelta
from norfab.core.worker import NFPWorker, Task, Job
from norfab.models import Result
from typing import Any, Union, List
from requests.packages.urllib3.exceptions import InsecureRequestWarning
from .netbox_models import NetboxFastApiArgs
from diskcache import FanoutCache

from .design_tasks import NetboxDesignTasks
from .interfaces_tasks import NetboxInterfacesTasks
from .devices_tasks import NetboxDevicesTasks
from .connections_tasks import NetboxConnectionsTasks
from .netbox_exceptions import UnsupportedNetboxVersion
from .circuits_tasks import NetboxCircuitsTasks
from .nornir_inventory_tasks import NetboxNornirInventoryTasks
from .bgp_peerings_tasks import NetboxBgpPeeringsTasks
from .prefix_tasks import NetboxPrefixTasks
from .containerlab_inventory_tasks import NetboxContainerlabInventoryTasks
from .ip_tasks import NetboxIpTasks
from .branch_tasks import NetboxBranchTasks

SERVICE = "netbox"

log = logging.getLogger(__name__)


# ----------------------------------------------------------------------
# HELPER FUNCTIONS
# ----------------------------------------------------------------------


def _form_query_v4(obj, filters, fields, alias=None) -> str:
    """
    Helper function to form graphql query for Netbox version 4.

    Args:
        obj (str): The object to return data for, e.g., 'device', 'interface', 'ip_address'.
        filters (dict): A dictionary of key-value pairs to filter by.
        fields (list): A list of data fields to return.
        alias (str, optional): An alias value for the requested object.

    Returns:
        str: A formatted GraphQL query string.
    """
    filters_list = []
    fields = " ".join(fields)
    if isinstance(filters, str):
        filters = filters.replace("'", '"')  # swap quotes
        if alias:
            query = f"{alias}: {obj}(filters: {filters}) {{{fields}}}"
        else:
            query = f"{obj}(filters: {filters}) {{{fields}}}"
    elif isinstance(filters, dict):
        for k, v in filters.items():
            if isinstance(v, (list, set, tuple)):
                items = ", ".join(f'"{i}"' for i in v)
                filters_list.append(f"{k}: [{items}]")
            elif "{" in v and "}" in v:
                filters_list.append(f"{k}: {v}")
            else:
                filters_list.append(f'{k}: "{v}"')
        filters_string = ", ".join(filters_list)
        filters_string = filters_string.replace("'", '"')  # swap quotes
        if alias:
            query = f"{alias}: {obj}(filters: {{{filters_string}}}) {{{fields}}}"
        else:
            query = f"{obj}(filters: {{{filters_string}}}) {{{fields}}}"

    return query


class NetboxWorker(
    NFPWorker,
    NetboxDesignTasks,
    NetboxInterfacesTasks,
    NetboxDevicesTasks,
    NetboxConnectionsTasks,
    NetboxCircuitsTasks,
    NetboxNornirInventoryTasks,
    NetboxBgpPeeringsTasks,
    NetboxPrefixTasks,
    NetboxContainerlabInventoryTasks,
    NetboxIpTasks,
    NetboxBranchTasks,
):
    """
    NetboxWorker class for interacting with Netbox API and managing inventory.

    Args:
        inventory (dict): The inventory data.
        broker (object): The broker instance.
        worker_name (str): The name of the worker.
        exit_event (threading.Event, optional): Event to signal exit.
        init_done_event (threading.Event, optional): Event to signal initialization completion.
        log_level (int, optional): Logging level.
        log_queue (object, optional): Queue for logging.

    Raises:
        AssertionError: If the inventory has no Netbox instances.

    Attributes:
        default_instance (str): Default Netbox instance name.
        inventory (dict): Inventory data.
        nb_version (tuple): Netbox version.
        compatible_ge_v4 (tuple): Minimum supported Netbox v4 version (4.4.0+).
    """

    default_instance = None
    inventory = None
    nb_version = {}  # dict keyed by instance name and version
    compatible_ge_v4 = (
        4,
        4,
        0,
    )  # 4.4.0 - minimum supported Netbox v4

    def __init__(
        self,
        inventory,
        broker,
        worker_name,
        exit_event=None,
        init_done_event=None,
        log_level=None,
        log_queue: object = None,
    ):
        super().__init__(
            inventory, broker, SERVICE, worker_name, exit_event, log_level, log_queue
        )
        self.init_done_event = init_done_event
        self.cache = None

        # get inventory from broker
        self.netbox_inventory = self.load_inventory()
        if not self.netbox_inventory:
            log.critical(
                f"{self.name} - Broker {self.broker} returned no inventory for {self.name}, killing myself..."
            )
            self.destroy()

        assert self.netbox_inventory.get(
            "instances"
        ), f"{self.name} - inventory has no Netbox instances"

        # extract parameters from imvemtory
        self.netbox_connect_timeout = self.netbox_inventory.get(
            "netbox_connect_timeout", 10
        )
        self.netbox_read_timeout = self.netbox_inventory.get("netbox_read_timeout", 300)
        self.cache_use = self.netbox_inventory.get("cache_use", True)
        self.cache_ttl = self.netbox_inventory.get("cache_ttl", 31557600)  # 1 Year
        self.branch_create_timeout = self.netbox_inventory.get(
            "branch_create_timeout", 120
        )

        # find default instance
        for name, params in self.netbox_inventory["instances"].items():
            if params.get("default") is True:
                self.default_instance = name
                break
        else:
            self.default_instance = name

        # check Netbox compatibility
        self._verify_compatibility()

        # instantiate cache
        self.cache_dir = os.path.join(self.base_dir, "cache")
        os.makedirs(self.cache_dir, exist_ok=True)
        self.cache = self._get_diskcache()

        self.init_done_event.set()
        log.info(f"{self.name} - Started")

    def worker_exit(self) -> None:
        """
        Worker exist sanity checks. Closes the cache if it exists.

        This method checks if the cache attribute is present and not None.
        If the cache exists, it closes the cache to release any resources
        associated with it.
        """
        if self.cache:
            self.cache.close()

    # ----------------------------------------------------------------------
    # Netbox Service Functions that exposed for calling
    # ----------------------------------------------------------------------

    @Task(fastapi={"methods": ["GET"], "schema": NetboxFastApiArgs.model_json_schema()})
    def get_inventory(self) -> Result:
        """
        NorFab Task to return running inventory for NetBox worker.

        Returns:
            dict: A dictionary containing the NetBox inventory.
        """
        return Result(
            task=f"{self.name}:get_inventory", result=dict(self.netbox_inventory)
        )

    @Task(fastapi={"methods": ["GET"], "schema": NetboxFastApiArgs.model_json_schema()})
    def get_version(self, **kwargs: Any) -> Result:
        """
        Retrieves the version information of Netbox instances.

        Returns:
            dict: A dictionary containing the version information of the Netbox
        """
        libs = {
            "norfab": "",
            "pynetbox": "",
            "requests": "",
            "python": sys.version.split(" ")[0],
            "platform": sys.platform,
            "diskcache": "",
            "netbox_version": self.nb_version,
        }
        # get version of packages installed
        for pkg in libs.keys():
            try:
                libs[pkg] = importlib.metadata.version(pkg)
            except importlib.metadata.PackageNotFoundError:
                pass

        return Result(task=f"{self.name}:get_version", result=libs)

    @Task(fastapi={"methods": ["GET"], "schema": NetboxFastApiArgs.model_json_schema()})
    def get_netbox_status(self, instance: Union[None, str] = None) -> Result:
        """
        Retrieve the status of NetBox instances.

        This method queries the status of a specific NetBox instance if the
        `instance` parameter is provided. If no instance is specified, it
        queries the status of all instances in the NetBox inventory.

        Args:
            instance (str, optional): The name of the specific NetBox instance to query.

        Returns:
            dict: A dictionary containing the status of the requested NetBox
                  instance(s).
        """
        ret = Result(result={}, task=f"{self.name}:get_netbox_status")
        if instance:
            ret.result[instance] = self._query_netbox_status(instance)
        else:
            for name in self.netbox_inventory["instances"].keys():
                ret.result[name] = self._query_netbox_status(name)
        return ret

    @Task(fastapi={"methods": ["GET"], "schema": NetboxFastApiArgs.model_json_schema()})
    def get_compatibility(self, job: Job) -> Result:
        """
        Checks the compatibility of Netbox instances based on their version.

        This method retrieves the status and version of Netbox instances and determines
        if they are compatible with the required versions. It logs a warning if any
        instance is not reachable.

        Args:
            job: NorFab Job object containing relevant metadata

        Returns:
            dict: A dictionary where the keys are the instance names and the values are
                  booleans indicating compatibility (True/False) or None if the instance
                  is not reachable.
        """
        ret = Result(task=f"{self.name}:get_compatibility", result={})
        netbox_status = self.get_netbox_status(job=job)
        for instance, params in netbox_status.result.items():
            if params["status"] is not True:
                log.warning(f"{self.name} - {instance} Netbox instance not reachable")
                ret.result[instance] = None
            else:
                if "-docker-" in params["netbox-version"].lower():
                    self.nb_version[instance] = tuple(
                        [
                            int(i)
                            for i in params["netbox-version"]
                            .lower()
                            .split("-docker-")[0]
                            .split(".")
                        ]
                    )
                else:
                    self.nb_version[instance] = tuple(
                        [int(i) for i in params["netbox-version"].split(".")]
                    )
                # check Netbox 4.4+ compatibility
                if self.nb_version[instance] >= self.compatible_ge_v4:
                    ret.result[instance] = True
                else:
                    ret.result[instance] = False
                    log.error(
                        f"{self.name} - {instance} Netbox version {self.nb_version[instance]} is not supported, "
                        f"minimum required version is {self.compatible_ge_v4}"
                    )

        return ret

    def _verify_compatibility(self):
        """
        Verifies the compatibility of Netbox instances.

        This method checks the compatibility of Netbox instances by calling the
        `get_compatibility` method. If any of the instances are not compatible,
        it raises a RuntimeError with a message indicating which instances are
        not compatible.

        Raises:
            RuntimeError: If any of the Netbox instances are not compatible.
        """
        compatibility = self.get_compatibility(job=Job())
        if not all(i is not False for i in compatibility.result.values()):
            raise RuntimeError(
                f"{self.name} - not all Netbox instances are compatible: {compatibility.result}"
            )

    def has_plugin(self, plugin_name: str, instance: str, strict: bool = False) -> bool:
        """
        Check if a specified plugin is installed in a given NetBox instance.

        Args:
            plugin_name (str): The name of the plugin to check for.
            instance (str): The identifier or address of the NetBox instance.
            strict (bool, optional): If True, raises a RuntimeError when the plugin is not found.

        Returns:
            bool: True if the plugin is installed, False otherwise.
        """
        nb_status = self._query_netbox_status(instance)

        if plugin_name in nb_status["plugins"]:
            return True
        elif strict is True:
            raise RuntimeError(
                f"'{instance}' Netbox instance has no '{plugin_name}' plugin installed"
            )

        return False

    def _query_netbox_status(self, name):
        """
        Queries the Netbox API for the status of a given instance.

        Args:
            name (str): The name of the Netbox instance to query.

        Returns:
            dict: A dictionary containing the status and any error message. The dictionary has the following keys:

                - "error" (str or None): Error message if the query failed, otherwise None.
                - "status" (bool): True if the query was successful, False otherwise.
                - Additional keys from the Netbox API response if the query was successful.

        Raises:
            None: All exceptions are caught and handled within the method.
        """
        params = self._get_instance_params(name)

        ret = {
            "error": None,
            "status": True,
        }

        try:
            response = requests.get(
                f"{params['url']}/api/status",
                verify=params.get("ssl_verify", True),
                timeout=(self.netbox_connect_timeout, self.netbox_read_timeout),
                headers={
                    "Content-Type": "application/json",
                    "Accept": "application/json",
                    "Authorization": f"Token {params['token']}",
                },
            )
            response.raise_for_status()
            ret.update(response.json())
        except Exception as e:
            ret["status"] = False
            msg = (
                f"{self.name} - failed to query Netbox API URL "
                f"'{params['url']}', token ends "
                f"with '..{params['token'][-6:]}'; error: '{e}'"
            )
            log.error(msg)
            ret["error"] = msg

        return ret

    def _get_instance_params(self, name: str = None) -> dict:
        """
        Retrieve instance parameters from the NetBox inventory.

        Args:
            name (str): The name of the instance to retrieve parameters for.

        Returns:
            dict: A dictionary containing the parameters of the specified instance.

        Raises:
            KeyError: If the specified instance name is not found in the inventory.

        If the `ssl_verify` parameter is set to False, SSL warnings will be disabled.
        Otherwise, SSL warnings will be enabled.
        """
        name = name or self.default_instance
        params = self.netbox_inventory["instances"][name]

        # check if need to disable SSL warnings
        if params.get("ssl_verify") == False:
            requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
        else:
            requests.packages.urllib3.enable_warnings(InsecureRequestWarning)

        return params

    def _get_pynetbox(self, instance, branch: str = None):
        """
        Helper function to instantiate a pynetbox API object.

        Args:
            instance (str): The instance name for which to get the pynetbox API object.
            branch (str, optional): Branch name to use, need to have branching plugin installed.
                Creates branch if it does not exist in Netbox.

        Returns:
            pynetbox.core.api.Api: An instantiated pynetbox API object.

        Raises:
            Exception: If the pynetbox library is not installed.

        If SSL verification is disabled in the instance parameters,
        this function will disable warnings for insecure requests.
        """
        params = self._get_instance_params(instance)
        nb = pynetbox.api(url=params["url"], token=params["token"], threading=True)

        if params.get("ssl_verify") == False:
            nb.http_session.verify = False

        # add branch
        if branch is not None and self.has_plugin(
            "netbox_branching", instance, strict=True
        ):
            try:
                nb_branch = nb.plugins.branching.branches.get(name=branch)
            except Exception:
                msg = "Failed to retrieve branch '{branch}' from Netbox"
                raise RuntimeError(msg)

            # create new branch
            if not nb_branch:
                nb_branch = nb.plugins.branching.branches.create(name=branch)

            # wait for branch provisioning to complete
            if not nb_branch.status.value.lower() == "ready":
                retries = 0
                while retries < self.branch_create_timeout:
                    nb_branch = nb.plugins.branching.branches.get(name=branch)
                    if nb_branch.status.value.lower() == "ready":
                        break
                    time.sleep(1)
                    retries += 1
                else:
                    raise RuntimeError(f"Branch '{branch}' was created but not ready")

            nb.http_session.headers["X-NetBox-Branch"] = nb_branch.schema_id

            log.info(f"Instantiated pynetbox instance with branch '{branch}'")

        return nb

    def _get_diskcache(self) -> FanoutCache:
        """
        Creates and returns a FanoutCache object.

        The FanoutCache is configured with the specified directory, number of shards,
        timeout, and size limit.

        Returns:
            FanoutCache: A configured FanoutCache instance.
        """
        return FanoutCache(
            directory=self.cache_dir,
            shards=4,
            timeout=1,  # 1 second
            size_limit=1073741824,  #  GigaByte
        )

    @Task(fastapi={"methods": ["GET"], "schema": NetboxFastApiArgs.model_json_schema()})
    def cache_list(self, keys: str = "*", details: bool = False) -> Result:
        """
        Retrieve a list of cache keys, optionally with details about each key.

        Args:
            keys (str): A pattern to match cache keys against. Defaults to "*".
            details (bool): If True, include detailed information about each cache key. Defaults to False.

        Returns:
            list: A list of cache keys or a list of dictionaries with detailed information if `details` is True.
        """
        self.cache.expire()
        ret = Result(task=f"{self.name}:cache_list", result=[])
        for cache_key in self.cache:
            if fnmatchcase(cache_key, keys):
                if details:
                    _, expires = self.cache.get(cache_key, expire_time=True)
                    expires = datetime.fromtimestamp(expires)
                    creation = expires - timedelta(seconds=self.cache_ttl)
                    age = datetime.now() - creation
                    ret.result.append(
                        {
                            "key": cache_key,
                            "age": str(age),
                            "creation": str(creation),
                            "expires": str(expires),
                        }
                    )
                else:
                    ret.result.append(cache_key)
        return ret

    @Task(
        fastapi={"methods": ["DELETE"], "schema": NetboxFastApiArgs.model_json_schema()}
    )
    def cache_clear(self, job: Job, key: str = None, keys: str = None) -> Result:
        """
        Clears specified keys from the cache.

        Args:
            job: NorFab Job object containing relevant metadata
            key (str, optional): A specific key to remove from the cache.
            keys (str, optional): A glob pattern to match multiple keys to remove from the cache.

        Returns:
            list: A list of keys that were successfully removed from the cache.

        Raises:
            RuntimeError: If a specified key or a key matching the glob pattern could not be removed from the cache.

        Notes:

        - If neither `key` nor `keys` is provided, the function will return a message indicating that there is nothing to clear.
        - If `key` is provided, it will attempt to remove that specific key from the cache.
        - If `keys` is provided, it will attempt to remove all keys matching the glob pattern from the cache.
        """
        ret = Result(task=f"{self.name}:cache_clear", result=[])
        # check if has keys to clear
        if key == keys == None:  # noqa
            ret.result = "Noting to clear, specify key or keys"
            return ret
        # remove specific key from cache
        if key:
            if key in self.cache:
                if self.cache.delete(key, retry=True):
                    ret.result.append(key)
                else:
                    raise RuntimeError(f"Failed to remove {key} from cache")
            else:
                ret.messages.append(f"Key {key} not in cache.")
        # remove all keys matching glob pattern
        if keys:
            for cache_key in self.cache:
                if fnmatchcase(cache_key, keys):
                    if self.cache.delete(cache_key, retry=True):
                        ret.result.append(cache_key)
                    else:
                        raise RuntimeError(f"Failed to remove {key} from cache")
        return ret

    @Task(fastapi={"methods": ["GET"], "schema": NetboxFastApiArgs.model_json_schema()})
    def cache_get(
        self, job: Job, key: str = None, keys: str = None, raise_missing: bool = False
    ) -> Result:
        """
        Retrieve values from the cache based on a specific key or a pattern of keys.

        Args:
            job: NorFab Job object containing relevant metadata
            key (str, optional): A specific key to retrieve from the cache.
            keys (str, optional): A glob pattern to match multiple keys in the cache.
            raise_missing (bool, optional): If True, raises a KeyError if the specific
                key is not found in the cache. Defaults to False.

        Returns:
            dict: A dictionary containing the results of the cache retrieval. The keys are
                the cache keys and the values are the corresponding cache values.

        Raises:
            KeyError: If raise_missing is True and the specific key is not found in the cache.
        """
        ret = Result(task=f"{self.name}:cache_clear", result={})
        # get specific key from cache
        if key:
            if key in self.cache:
                ret.result[key] = self.cache[key]
            elif raise_missing:
                raise KeyError(f"Key {key} not in cache.")
        # get all keys matching glob pattern
        if keys:
            for cache_key in self.cache:
                if fnmatchcase(cache_key, keys):
                    ret.result[cache_key] = self.cache[cache_key]
        return ret

    @Task(
        fastapi={"methods": ["POST"], "schema": NetboxFastApiArgs.model_json_schema()}
    )
    def graphql(
        self,
        job: Job,
        instance: Union[None, str] = None,
        dry_run: bool = False,
        obj: Union[str, dict] = None,
        filters: Union[None, dict, str] = None,
        fields: Union[None, list] = None,
        queries: Union[None, dict] = None,
        query_string: str = None,
    ) -> Result:
        """
        Function to query Netbox v3 or Netbox v4 GraphQL API.

        Args:
            job: NorFab Job object containing relevant metadata
            instance: Netbox instance name
            dry_run: only return query content, do not run it
            obj: Object to query
            filters: Filters to apply to the query
            fields: Fields to retrieve in the query
            queries: Dictionary of queries to execute
            query_string: Raw query string to execute

        Returns:
            dict: GraphQL request data returned by Netbox

        Raises:
            RuntimeError: If required arguments are not provided
            Exception: If GraphQL query fails
        """
        nb_params = self._get_instance_params(instance)
        instance = instance or self.default_instance
        ret = Result(task=f"{self.name}:graphql", resources=[instance])

        # form graphql query(ies) payload
        if queries:
            queries_list = []
            for alias, query_data in queries.items():
                query_data["alias"] = alias
                if self.nb_version[instance] >= (4, 4, 0):
                    queries_list.append(_form_query_v4(**query_data))
                else:
                    raise UnsupportedNetboxVersion(
                        f"{self.name} - Netbox version {self.nb_version[instance]} is not supported, "
                        f"minimum required version is {self.compatible_ge_v4}"
                    )
            queries_strings = "    ".join(queries_list)
            query = f"query {{{queries_strings}}}"
        elif obj and filters and fields:
            if self.nb_version[instance] >= (4, 4, 0):
                query = _form_query_v4(obj, filters, fields)
            else:
                raise UnsupportedNetboxVersion(
                    f"{self.name} - Netbox version {self.nb_version[instance]} is not supported, "
                    f"minimum required version is {self.compatible_ge_v4}"
                )
            query = f"query {{{query}}}"
        elif query_string:
            query = query_string
        else:
            raise RuntimeError(
                f"{self.name} - graphql method expects queries argument or obj, filters, "
                f"fields arguments or query_string argument provided"
            )
        payload = json.dumps({"query": query})

        # form and return dry run response
        if dry_run:
            ret.result = {
                "url": f"{nb_params['url']}/graphql/",
                "data": payload,
                "verify": nb_params.get("ssl_verify", True),
                "headers": {
                    "Content-Type": "application/json",
                    "Accept": "application/json",
                    "Authorization": f"Token ...{nb_params['token'][-6:]}",
                },
            }
            return ret

        # send request to Netbox GraphQL API
        log.debug(
            f"{self.name} - sending GraphQL query '{payload}' to URL '{nb_params['url']}/graphql/'"
        )
        req = requests.post(
            url=f"{nb_params['url']}/graphql/",
            headers={
                "Content-Type": "application/json",
                "Accept": "application/json",
                "Authorization": f"Token {nb_params['token']}",
            },
            data=payload,
            verify=nb_params.get("ssl_verify", True),
            timeout=(self.netbox_connect_timeout, self.netbox_read_timeout),
        )
        try:
            req.raise_for_status()
        except Exception:
            raise Exception(
                f"{self.name} -  Netbox GraphQL query failed, query '{query}', "
                f"URL '{req.url}', status-code '{req.status_code}', reason '{req.reason}', "
                f"response content '{req.text}'"
            )

        # return results
        reply = req.json()
        if reply.get("errors"):
            msg = f"{self.name} - GrapQL query error '{reply['errors']}', query '{payload}'"
            log.error(msg)
            ret.errors.append(msg)
            if reply.get("data"):
                ret.result = reply["data"]  # at least return some data
        elif queries or query_string:
            ret.result = reply["data"]
        else:
            ret.result = reply["data"][obj]

        return ret

    @Task(
        fastapi={"methods": ["POST"], "schema": NetboxFastApiArgs.model_json_schema()}
    )
    def rest(
        self,
        job: Job,
        instance: Union[None, str] = None,
        method: str = "get",
        api: str = "",
        **kwargs: Any,
    ) -> Result:
        """
        Sends a request to the Netbox REST API.

        Args:
            instance (str, optional): The Netbox instance name to get parameters for.
            method (str, optional): The HTTP method to use for the request (e.g., 'get', 'post'). Defaults to "get".
            api (str, optional): The API endpoint to send the request to. Defaults to "".
            **kwargs: Additional arguments to pass to the request (e.g., params, data, json).

        Returns:
            Union[dict, list]: The JSON response from the API, parsed into a dictionary or list.

        Raises:
            requests.exceptions.HTTPError: If the HTTP request returned an unsuccessful status code.
        """
        ret = Result(task=f"{self.name}:rest", result={})
        nb_params = self._get_instance_params(instance)
        api = api.strip("/")

        # send request to Netbox REST API
        response = getattr(requests, method)(
            url=f"{nb_params['url']}/api/{api}/",
            headers={
                "Content-Type": "application/json",
                "Accept": "application/json",
                "Authorization": f"Token {nb_params['token']}",
            },
            verify=nb_params.get("ssl_verify", True),
            **kwargs,
        )

        try:
            response.raise_for_status()
        except Exception as e:
            log.error(f"REST API request failed, error: {e}")
            ret.result = response.status_code
            return ret

        try:
            ret.result = response.json()
        except Exception as e:
            log.debug(f"Failed to decode json, error: {e}")
            ret.result = response.text if response.text else response.status_code

        return ret

    def expand_alphanumeric_range(self, range_pattern: str) -> list:
        """
        Expand alphanumeric ranges.

        Examples:
            - Ethernet[1-3] -> ['Ethernet1', 'Ethernet2', 'Ethernet3']
            - [ge,xe]-0/0/[0-9] -> ['ge-0/0/0', 'ge-0/0/1', ..., 'xe-0/0/9']
        """
        # Find all bracketed patterns
        bracket_pattern = r"\[([^\]]+)\]"
        matches = list(re.finditer(bracket_pattern, range_pattern))

        if not matches:
            # No ranges found, return as-is
            return [range_pattern]

        # Start with a single template
        templates = [range_pattern]

        # Process each bracket from left to right
        for match in matches:
            bracket_content = match.group(1)
            new_templates = []

            # Check if it's a comma-separated list
            if "," in bracket_content:
                options = [opt.strip() for opt in bracket_content.split(",")]
                for template in templates:
                    for option in options:
                        new_templates.append(
                            template.replace(f"[{bracket_content}]", option, 1)
                        )

            # Check if it's a numeric range
            elif (
                "-" in bracket_content
                and bracket_content.replace("-", "").replace(" ", "").isdigit()
            ):
                parts = bracket_content.split("-")
                if len(parts) == 2:
                    try:
                        start = int(parts[0].strip())
                        end = int(parts[1].strip())
                        for template in templates:
                            for num in range(start, end + 1):
                                new_templates.append(
                                    template.replace(
                                        f"[{bracket_content}]", str(num), 1
                                    )
                                )
                    except ValueError:
                        # If conversion fails, treat as literal
                        for template in templates:
                            new_templates.append(
                                template.replace(
                                    f"[{bracket_content}]", bracket_content, 1
                                )
                            )
            else:
                # Treat as literal
                for template in templates:
                    new_templates.append(
                        template.replace(f"[{bracket_content}]", bracket_content, 1)
                    )

            templates = new_templates

        return templates

    def compare_netbox_object_state(
        self,
        desired_state: dict,
        current_state: dict,
        ignore_fields: Union[list, None] = None,
        ignore_if_not_empty: Union[list, None] = None,
        diff: dict = None,
    ) -> tuple:
        """
        Compare desired state with current NetBox object state and return fields that need updating.

        Args:
            desired_state (dict): Dictionary with desired field values.
            current_state (dict): Dictionary with current NetBox object field values.
            ignore_fields (list, optional): List of field names to ignore completely.
            ignore_if_not_empty (list, optional): List of field names to ignore if they have
                non-empty values in current_state (won't overwrite existing data).
            diff (dict, optional): Dictionary to accumulate field differences. If not provided,
                a new dictionary will be created.

        Returns:
            tuple: A tuple containing:
                - updates (dict): Dictionary containing only fields that need to be updated with their new values.
                - diff (dict): Dictionary containing the differences with '+' (new value) and '-' (old value) keys.

        Example:
            >>> desired = {"serial": "ABC123", "asset_tag": "TAG001", "comments": "New comment"}
            >>> current = {"serial": "OLD123", "asset_tag": "", "comments": "Existing"}
            >>> ignore_fields = []
            >>> ignore_if_not_empty = ["comments"]
            >>> updates, diff = compare_netbox_object_state(desired, current, ignore_fields, ignore_if_not_empty)
            >>> updates
            {"serial": "ABC123", "asset_tag": "TAG001"}
        """
        ignore_fields = ignore_fields or []
        ignore_if_not_empty = ignore_if_not_empty or []
        updates = {}
        diff = diff or {}

        for field, desired_value in desired_state.items():
            # Skip if field is in ignore list
            if field in ignore_fields:
                continue

            # Get current value, default to None if field doesn't exist
            current_value = current_state.get(field)

            # Skip if field is in ignore_if_not_empty and current value is not empty
            if field in ignore_if_not_empty and current_value:
                continue

            # Compare values and add to updates if different
            if current_value != desired_value:
                updates[field] = desired_value
                diff[field] = {
                    "-": current_value,
                    "+": desired_value,
                }

        return updates, diff

    def get_nornir_hosts(self, kwargs: dict, timeout: int) -> List[str]:
        """
        Retrieves a list of unique Nornir hosts from Nornir service based on provided filter criteria.

        Args:
            kwargs (dict): Dictionary of keyword arguments, where keys starting with 'F' are used as filters.
            timeout (int): Timeout value (in seconds) for the job execution.

        Returns:
            list: Sorted list of unique Nornir host names that match the filter criteria.

        Notes:
            - Only filters with keys starting with 'F' are considered.
            - Hosts are collected from all workers where the job did not fail.
        """
        ret = []
        filters = {k: v for k, v in kwargs.items() if k.startswith("F")}
        if filters:
            nornir_hosts = self.client.run_job(
                "nornir",
                "get_nornir_hosts",
                kwargs=filters,
                workers="all",
                timeout=timeout,
            )
            for w, r in nornir_hosts.items():
                if r["failed"] is False and isinstance(r["result"], list):
                    ret.extend(r["result"])

        return list(sorted(set(ret)))



import logging
import builtins

from picle.models import Outputters, PipeFunctionsModel
from pydantic import (
    StrictBool,
    StrictStr,
    Field,
)
from typing import Union, Optional, List
from ..common import log_error_or_result, listen_events
from .netbox_picle_shell_common import NetboxClientRunJobArgs
from norfab.workers.netbox_worker.netbox_models import NetboxCommonArgs

log = logging.getLogger(__name__)


class GetInterfaces(NetboxCommonArgs, NetboxClientRunJobArgs):
    devices: Union[StrictStr, List] = Field(
        ..., description="Devices to retrieve interface for"
    )
    ip_addresses: Optional[StrictBool] = Field(
        None,
        description="Retrieves interface IP addresses",
        json_schema_extra={"presence": True},
        alias="ip-addresses",
    )
    inventory_items: Optional[StrictBool] = Field(
        None,
        description="Retrieves interface inventory items",
        json_schema_extra={"presence": True},
        alias="inventory-items",
    )
    dry_run: Optional[StrictBool] = Field(
        None,
        description="Only return query content, do not run it",
        json_schema_extra={"presence": True},
        alias="dry-run",
    )
    interface_regex: StrictStr = Field(
        None,
        description="Regex pattern to match interfaces and ports",
        alias="interface-regex",
    )
    interface_list: Union[StrictStr, List[StrictStr]] = Field(
        None,
        description="List of interfaces names to filter",
        alias="interface-list",
    )

    @staticmethod
    @listen_events
    def run(uuid, *args, **kwargs):
        NFCLIENT = builtins.NFCLIENT
        workers = kwargs.pop("workers", "any")
        timeout = kwargs.pop("timeout", 600)
        verbose_result = kwargs.pop("verbose_result", False)
        nowait = kwargs.pop("nowait", False)
        if isinstance(kwargs["devices"], str):
            kwargs["devices"] = [kwargs["devices"]]
        if isinstance(kwargs.get("interface_list"), str):
            kwargs["interface_list"] = [kwargs["interface_list"]]
        result = NFCLIENT.run_job(
            "netbox",
            "get_interfaces",
            workers=workers,
            args=args,
            kwargs=kwargs,
            timeout=timeout,
            uuid=uuid,
            nowait=nowait,
        )

        if nowait:
            return result, Outputters.outputter_nested

        result = log_error_or_result(result, verbose_result=verbose_result)
        return result

    class PicleConfig:
        outputter = Outputters.outputter_json
        pipe = PipeFunctionsModel



"""
Pydantic models for Python logging.config.dictConfig() configuration schema.

Reference: https://docs.python.org/3/library/logging.config.html#logging-config-dictschema

The top-level model is LoggingConfig, which is a full, typed representation of
the dictConfig dictionary schema including the NorFab-specific log_events extension.
"""

from pydantic import BaseModel, StrictBool, StrictInt, StrictStr, Field, ConfigDict
from typing import Literal, Union, List, Any, Dict

# ------------------------------------------------------
# Formatter models
# ------------------------------------------------------


class LoggingFormatterConfig(BaseModel):
    """
    Configuration for a single Python logging formatter (dictConfig formatters entry).

    Corresponds to arguments passed to logging.Formatter.__init__() plus an
    optional class key for custom formatter subclasses, or a () key for a
    fully user-defined factory callable.

    Reference: https://docs.python.org/3/library/logging.config.html#dictionary-schema-details
    """

    model_config = ConfigDict(extra="allow", populate_by_name=True)

    class_: StrictStr = Field(
        None,
        alias="class",
        description=(
            "Dotted path to a custom Formatter subclass. "
            "Omit to use the default logging.Formatter."
        ),
    )
    format: StrictStr = Field(
        None,
        description="Log record format string.",
    )
    datefmt: StrictStr = Field(
        None,
        description="Date/time format string passed to time.strftime().",
    )
    style: Literal["%", "{", "$"] = Field(
        None,
        description=(
            "Format string style. "
            "'%' (printf, default), '{' (str.format), or '$' (string.Template)."
        ),
    )
    validate_: StrictBool = Field(
        None,
        description=(
            "If True (default), validate the format string against the log record "
            "fields. Added in Python 3.8."
        ),
        examples=[True],
        alias="validate",
    )
    defaults: Dict[StrictStr, Any] = Field(
        None,
        description=(
            "A dictionary of default values for custom formatting fields, "
            "available for use in the format string. Added in Python 3.12."
        ),
    )


# ------------------------------------------------------
# Filter models
# ------------------------------------------------------


class LoggingFilterConfig(BaseModel):
    """
    Configuration for a single Python logging filter (dictConfig filters entry).

    For the built-in logging.Filter provide name (logger name prefix).
    For user-defined filters supply a () factory key (extra fields allowed).

    Reference: https://docs.python.org/3/library/logging.config.html#dictionary-schema-details
    """

    model_config = ConfigDict(extra="allow", populate_by_name=True)

    name: StrictStr = Field(
        None,
        description=(
            "Logger name prefix for the built-in logging.Filter. "
            "Only records from loggers whose name starts with this value are passed through. "
            "Empty string (default) allows every record."
        ),
    )


# ------------------------------------------------------
# Handler models
# ------------------------------------------------------


class LoggingHandlerConfig(BaseModel):
    """
    Configuration for a single Python logging handler (dictConfig handlers entry).

    class is mandatory. All unrecognised keys are forwarded as keyword arguments
    to the handler's constructor, so handler-specific parameters (filename,
    maxBytes, stream, …) are stored via extra="allow".

    Reference: https://docs.python.org/3/library/logging.config.html#dictionary-schema-details
    """

    model_config = ConfigDict(extra="allow", populate_by_name=True)

    class_: StrictStr = Field(
        ...,
        alias="class",
        description="Fully-qualified dotted path to the handler class (mandatory).",
    )
    level: StrictStr = Field(
        None,
        description="Minimum severity level for this handler.",
    )
    formatter: StrictStr = Field(
        None,
        description="Id of the formatter (key in formatters) to use with this handler.",
    )
    filters: List[StrictStr] = Field(
        None,
        description="List of filter ids (keys in filters) to attach to this handler.",
    )
    # ---- Common handler constructor kwargs (kept typed for IDE support) ----
    # StreamHandler
    stream: StrictStr = Field(
        None,
        description=(
            "Stream for logging.StreamHandler. "
            "Use ext://sys.stdout or ext://sys.stderr."
        ),
    )
    # FileHandler / RotatingFileHandler / TimedRotatingFileHandler
    filename: StrictStr = Field(
        None,
        description="Absolute or relative path to the log file.",
    )
    mode: StrictStr = Field(
        None,
        description="File open mode for file-based handlers.",
    )
    encoding: StrictStr = Field(
        None,
        description="File encoding for file-based handlers.",
    )
    delay: StrictBool = Field(
        None,
        description=(
            "If True, file creation is deferred until the first record is emitted. "
            "Applies to FileHandler and subclasses."
        ),
    )
    # RotatingFileHandler
    maxBytes: StrictInt = Field(
        None,
        description=(
            "Maximum log file size in bytes before rollover "
            "(RotatingFileHandler). 0 disables rollover."
        ),
    )
    backupCount: StrictInt = Field(
        None,
        description=(
            "Number of backup files to keep after rollover "
            "(RotatingFileHandler / TimedRotatingFileHandler)."
        ),
    )
    # TimedRotatingFileHandler
    when: StrictStr = Field(
        None,
        description=(
            "Rollover interval type for TimedRotatingFileHandler. "
            "One of: 'S', 'M', 'H', 'D', 'W0'–'W6', 'midnight'."
        ),
    )
    interval: StrictInt = Field(
        None,
        description="Rollover interval value for TimedRotatingFileHandler.",
    )
    utc: StrictBool = Field(
        None,
        description=("Use UTC for rollover timing in TimedRotatingFileHandler."),
    )
    # SysLogHandler
    address: Union[StrictStr, List[Any]] = Field(
        None,
        description="Address for SysLogHandler, e.g. '/dev/log' or ['host', port].",
    )
    facility: StrictInt = Field(
        None,
        description="Syslog facility code for SysLogHandler.",
    )
    socktype: StrictInt = Field(
        None,
        description=(
            "Socket type for SysLogHandler: socket.SOCK_DGRAM (2) or "
            "socket.SOCK_STREAM (1)."
        ),
    )
    # SMTPHandler
    mailhost: Union[StrictStr, List[Any]] = Field(
        None,
        description="SMTP host (string) or [host, port] for SMTPHandler.",
    )
    fromaddr: StrictStr = Field(
        None,
        description="Sender address for SMTPHandler.",
    )
    toaddrs: List[StrictStr] = Field(
        None,
        description="Recipient address list for SMTPHandler.",
    )
    subject: StrictStr = Field(
        None,
        description="Email subject for SMTPHandler.",
    )
    # MemoryHandler
    capacity: StrictInt = Field(
        None,
        description="Buffer capacity (number of records) for MemoryHandler.",
    )
    flushLevel: StrictStr = Field(
        None,
        description="Level that triggers a flush for MemoryHandler.",
    )
    target: StrictStr = Field(
        None,
        description="Target handler id for MemoryHandler (cfg://handlers.<id>).",
    )
    # QueueHandler
    queue: StrictStr = Field(
        None,
        description=(
            "Queue factory callable path or cfg:// reference for QueueHandler. "
            "Defaults to an unbounded queue.Queue if omitted."
        ),
    )
    listener: StrictStr = Field(
        None,
        description=(
            "QueueListener subclass path for QueueHandler. "
            "Defaults to logging.handlers.QueueListener."
        ),
    )


# ------------------------------------------------------
# Logger models
# ------------------------------------------------------


class LoggingLoggerConfig(BaseModel):
    """
    Configuration for a named logger or the root logger (dictConfig loggers /
    root entry).

    Reference: https://docs.python.org/3/library/logging.config.html#dictionary-schema-details
    """

    level: StrictStr = Field(
        None,
        description="Logger severity level.",
    )
    handlers: List[StrictStr] = Field(
        None,
        description="List of handler ids (keys in handlers) to attach to this logger.",
    )
    filters: List[StrictStr] = Field(
        None,
        description="List of filter ids (keys in filters) to attach to this logger.",
    )
    propagate: StrictBool = Field(
        None,
        description=(
            "Whether records should propagate to ancestor loggers. "
            "Not applicable for the root logger."
        ),
    )


# ------------------------------------------------------
# Top-level logging configuration model
# ------------------------------------------------------


class LoggingConfig(BaseModel):
    """
    Full logging.config.dictConfig()-compatible configuration model.

    This model covers the complete dictConfig schema (version 1) and adds the
    NorFab-specific log_events flag.

    Reference: https://docs.python.org/3/library/logging.config.html#logging-config-dictschema

    Example (minimal)::

        {
            "version": 1,
            "disable_existing_loggers": false,
            "handlers": {
                "terminal": {"class": "logging.StreamHandler", "level": "WARNING"}
            },
            "root": {"level": "INFO", "handlers": ["terminal"]}
        }

    Example (full)::

        {
            "version": 1,
            "disable_existing_loggers": false,
            "incremental": false,
            "formatters": {
                "default": {
                    "class": "logging.Formatter",
                    "datefmt": "%Y-%m-%d %H:%M:%S",
                    "format": "%(asctime)s.%(msecs)d %(levelname)s [%(name)s:%(lineno)d ] -- %(message)s"
                }
            },
            "filters": {
                "norfab_only": {"name": "norfab"}
            },
            "handlers": {
                "terminal": {
                    "class": "logging.StreamHandler",
                    "formatter": "default",
                    "level": "WARNING",
                    "stream": "ext://sys.stderr"
                },
                "file": {
                    "class": "logging.handlers.RotatingFileHandler",
                    "formatter": "default",
                    "level": "INFO",
                    "filename": "/var/log/norfab/norfab.log",
                    "maxBytes": 1024000,
                    "backupCount": 50,
                    "encoding": "utf-8",
                    "mode": "a",
                    "delay": false
                }
            },
            "loggers": {
                "norfab.workers": {"level": "DEBUG", "propagate": true}
            },
            "root": {"level": "INFO", "handlers": ["terminal", "file"]},
            "log_events": true
        }
    """

    version: StrictInt = Field(
        1,
        description=(
            "dictConfig schema version. Must be 1 — the only valid value "
            "in the current Python standard library."
        ),
    )
    disable_existing_loggers: StrictBool = Field(
        True,
        description=(
            "If True (Python stdlib default), any loggers that exist at "
            "configuration time and are not explicitly named are disabled. "
            "Set to False to preserve existing loggers."
        ),
    )
    incremental: StrictBool = Field(
        False,
        description=(
            "If True, the configuration is applied incrementally — only the "
            "level and propagate settings of existing handlers and loggers "
            "are updated; formatters and filters are ignored."
        ),
    )
    formatters: Dict[StrictStr, LoggingFormatterConfig] = Field(
        None,
        description="Named formatter definitions keyed by formatter id.",
    )
    filters: Dict[StrictStr, LoggingFilterConfig] = Field(
        None,
        description="Named filter definitions keyed by filter id.",
    )
    handlers: Dict[StrictStr, LoggingHandlerConfig] = Field(
        None,
        description="Name of handler definitions",
    )
    loggers: Dict[StrictStr, LoggingLoggerConfig] = Field(
        None,
        description=(
            "Per-logger configurations keyed by logger name "
            "(as used in logging.getLogger(name))."
        ),
    )
    root: LoggingLoggerConfig = Field(
        None,
        description=(
            "Root logger configuration. propagate is not applicable here. "
            "All loggers without an explicit propagate=False ultimately forward "
            "records to the root logger."
        ),
    )
    # ------------------------------------------------------------------
    # NorFab extension — not part of the Python dictConfig standard
    # ------------------------------------------------------------------
    log_events: StrictBool = Field(
        None,
        description=(
            "NorFab-specific flag. When True, NorFab system events are forwarded "
            "to the Python logging system in addition to the internal event bus."
        ),
    )



import builtins

from pydantic import (
    BaseModel,
    StrictBool,
    StrictInt,
    StrictStr,
    Field,
)
from enum import Enum
from typing import Union, Optional, List, Dict
from norfab.models import NorFabClientRunJob

# --------------------------------------------------------------------------
# NETBOX WORKER CONFIGURATION MODEL
# --------------------------------------------------------------------------


class CacheUseEnum(str, Enum):
    force = "force"
    refresh = "refrresh"


class NetboxInstanceConfig(BaseModel):
    default: StrictBool = Field(
        None, description="Is this default instance of Netbox or not"
    )
    url: StrictStr = Field(None, description="Netbox URL")
    token: StrictStr = Field(None, description="Netbox auth token")
    ssl_verify: StrictBool = Field(True, description="Verify SSL vertsor not")


class NetboxConfigModel(BaseModel):
    cache_use: Union[CacheUseEnum, StrictBool] = Field(
        True, description="Use cache or not"
    )
    cache_ttl: StrictInt = Field(True, description="Cache TTL")
    instances: Dict[StrictStr, NetboxInstanceConfig] = Field(
        None, description="Netbox instance config keyed by instance name"
    )


# --------------------------------------------------------------------------
# CORE NETBOX WORKER MODELS
# --------------------------------------------------------------------------


class NetboxCommonArgs(BaseModel):
    """Model to enlist arguments common across Netbox service tasks"""

    instance: Optional[StrictStr] = Field(
        None,
        description="Netbox instance name to target",
    )
    dry_run: StrictBool = Field(
        None,
        description="Do not commit to database",
        alias="dry-run",
        json_schema_extra={"presence": True},
    )
    branch: Union[None, StrictStr] = Field(
        None, description="Branching plugin branch name to use"
    )

    @staticmethod
    def source_instance():
        NFCLIENT = builtins.NFCLIENT
        reply = NFCLIENT.run_job("netbox", "get_inventory", workers="any")
        for worker_name, inventory in reply.items():
            return list(inventory["result"]["instances"])


class NetboxFastApiArgs(NorFabClientRunJob):
    """Model to specify arguments for FastAPI REST API endpoints"""

    workers: Union[StrictStr, List[StrictStr]] = Field(
        "any", description="Filter worker to target"
    )


class InterfaceTypeEnum(str, Enum):
    virtual = "virtual"
    other = "other"
    bridge = "bridge"
    lag = "lag"


class CreateDeviceInterfacesInput(NetboxCommonArgs, use_enum_values=True):
    devices: List = Field(
        ...,
        description="List of device names or device objects to create interfaces for",
    )
    interface_name: Union[StrictStr, List[StrictStr]] = Field(
        ...,
        description="Name(s) of the interface(s) to create",
    )
    interface_type: InterfaceTypeEnum = Field(
        "other",
        description="Name(s) of the interface(s) to create",
        alias="interface-type",
    )
    description: Union[None, StrictStr] = Field(
        None, description="Interface description"
    )
    speed: StrictInt = Field(None, description="Interface speed in Kbit/s")
    mtu: StrictInt = Field(None, description="Maximum transmission unit size in bytes")



import logging
import builtins

from picle.models import PipeFunctionsModel, Outputters
from pydantic import (
    BaseModel,
    Field,
)
from ..common import ClientRunJobArgs, log_error_or_result
from .workflow_picle_shell_run import WorkflowRunShell

SERVICE = "workflow"
log = logging.getLogger(__name__)

# ---------------------------------------------------------------------------------------------
# WORKFLOW SERVICE SHELL SHOW COMMANDS MODELS
# ---------------------------------------------------------------------------------------------


class WorkflowShowInventoryModel(ClientRunJobArgs):
    class PicleConfig:
        outputter = Outputters.outputter_yaml
        pipe = PipeFunctionsModel

    @staticmethod
    def run(*args, **kwargs):
        NFCLIENT = builtins.NFCLIENT
        workers = kwargs.pop("workers", "all")
        timeout = kwargs.pop("timeout", 600)
        verbose_result = kwargs.pop("verbose_result", False)

        result = NFCLIENT.run_job(
            "workflow",
            "get_inventory",
            kwargs=kwargs,
            workers=workers,
            timeout=timeout,
        )
        return log_error_or_result(result, verbose_result=verbose_result)


class WorkflowShowVersionModel(ClientRunJobArgs):
    class PicleConfig:
        outputter = Outputters.outputter_yaml
        pipe = PipeFunctionsModel

    @staticmethod
    def run(*args, **kwargs):
        NFCLIENT = builtins.NFCLIENT
        workers = kwargs.pop("workers", "all")
        timeout = kwargs.pop("timeout", 600)
        verbose_result = kwargs.pop("verbose_result", False)

        result = NFCLIENT.run_job(
            "workflow",
            "get_version",
            kwargs=kwargs,
            workers=workers,
            timeout=timeout,
        )
        return log_error_or_result(result, verbose_result=verbose_result)


class WorkflowShowCommandsModel(BaseModel):
    inventory: WorkflowShowInventoryModel = Field(
        None,
        description="Show workflow workers inventory data",
    )
    version: WorkflowShowVersionModel = Field(
        None,
        description="Show workflow service workers version report",
    )


# ---------------------------------------------------------------------------------------------
# WORKFLOW SERVICE MAIN SHELL MODEL
# ---------------------------------------------------------------------------------------------


class WorkflowServiceCommands(BaseModel):
    run: WorkflowRunShell = Field(None, description="Run workflows")

    class PicleConfig:
        subshell = True
        prompt = "nf[workflow]#"
